from __future__ import annotations

import copy
import inspect
import logging
import pathlib
from pprint import pformat as pf
from typing import TYPE_CHECKING, ClassVar, Dict, Generator, List, Optional, Tuple, Type, Union

import pytest

from great_expectations.compatibility.pydantic import DirectoryPath, validate_arguments
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.partitioners import ColumnPartitioner
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import AbstractDataContext, FileDataContext
from great_expectations.data_context import get_context as get_gx_context
from great_expectations.datasource.fluent.batch_request import (
    BatchParameters,
    BatchRequest,
)
from great_expectations.datasource.fluent.config import GxConfig
from great_expectations.datasource.fluent.constants import (
    _FLUENT_DATASOURCES_KEY,
)
from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice
from great_expectations.datasource.fluent.interfaces import (
    DataAsset,
    Datasource,
)
from great_expectations.datasource.fluent.metadatasource import MetaDatasource
from great_expectations.datasource.fluent.sources import (
    DataSourceManager,
    TypeRegistrationError,
)
from great_expectations.execution_engine import ExecutionEngine

yaml = YAMLHandler()

if TYPE_CHECKING:
    from great_expectations.core.config_provider import _ConfigurationProvider
    from great_expectations.datasource.datasource_dict import DatasourceDict
    from great_expectations.datasource.fluent.interfaces import Batch


logger = logging.getLogger(__name__)


# Fake DataContext for testing
class DataContext:
    """
    This is just a scaffold for exploring and iterating on our fluent datasources.

    Use `great_expectations.get_context()` for a real DataContext.
    """

    _context: ClassVar[Optional[DataContext]] = None
    _config: ClassVar[Optional[GxConfig]] = None  # (kilo59) should this live  here?

    _datasources: Dict[str, Datasource]
    root_directory: Union[DirectoryPath, str, None]

    @classmethod
    def get_context(
        cls,
        context_root_dir: Optional[DirectoryPath] = None,
        _config_file: str = "config.yaml",  # for ease of use during POC
    ):
        if not cls._context:
            cls._context = DataContext(context_root_dir=context_root_dir)

        assert cls._context
        return cls._context

    @validate_arguments
    def __init__(self, context_root_dir: Optional[DirectoryPath] = None) -> None:
        self.root_directory = context_root_dir
        self._data_sources: DataSourceManager = DataSourceManager(self)  # type: ignore[arg-type] # FIXME CoP
        self._datasources: Dict[str, Datasource] = {}
        self.config_provider: _ConfigurationProvider | None = None
        logger.info(f"Available Factories - {self._data_sources.factories}")
        logger.debug(f"`type_lookup` mapping ->\n{pf(self._data_sources.type_lookup)}")

    @property
    def data_sources(self) -> DataSourceManager:
        return self._data_sources

    @property
    def datasources(self) -> DatasourceDict:
        return self._datasources  # type: ignore[return-value] # FIXME CoP

    def _add_fluent_datasource(self, datasource: Datasource) -> Datasource:
        self._datasources[datasource.name] = datasource
        return datasource

    def _update_fluent_datasource(self, datasource: Datasource) -> Datasource:
        self._datasources[datasource.name] = datasource
        return datasource

    def _save_project_config(self) -> None: ...


def get_context(context_root_dir: Optional[DirectoryPath] = None, **kwargs):
    """Experimental get_context placeholder function."""
    logger.info(f"Getting context {context_root_dir or ''}")
    context = DataContext.get_context(context_root_dir=context_root_dir, **kwargs)
    return context


class DummyDataAsset(DataAsset):
    """Minimal Concrete DataAsset Implementation"""

    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> BatchRequest:
        return BatchRequest("datasource_name", "data_asset_name", options or {})

    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        raise NotImplementedError

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        raise NotImplementedError


@pytest.fixture(scope="function")
def context_sources_cleanup() -> Generator[DataSourceManager, None, None]:
    """Return the sources object and reset types/factories on teardown"""
    try:
        # setup
        sources_copy = copy.deepcopy(DataSourceManager._DataSourceManager__crud_registry)  # type: ignore[attr-defined] # FIXME CoP
        type_lookup_copy = copy.deepcopy(DataSourceManager.type_lookup)
        sources = get_context().data_sources

        assert "add_datasource" not in sources.factories, (
            "Datasource base class should not be registered as a source factory"
        )

        yield sources
    finally:
        DataSourceManager._DataSourceManager__crud_registry = sources_copy  # type: ignore[attr-defined] # FIXME CoP
        DataSourceManager.type_lookup = type_lookup_copy


@pytest.fixture(scope="function")
def empty_sources(context_sources_cleanup) -> Generator[DataSourceManager, None, None]:
    DataSourceManager._DataSourceManager__crud_registry.clear()  # type: ignore[attr-defined] # FIXME CoP
    DataSourceManager.type_lookup.clear()
    assert not DataSourceManager.type_lookup
    yield context_sources_cleanup


class DummyExecutionEngine(ExecutionEngine):
    def get_batch_data_and_markers(self, batch_spec):  # type: ignore[explicit-override] # FIXME
        raise NotImplementedError


@pytest.mark.unit
class TestMetaDatasource:
    def test__new__only_registers_expected_number_of_datasources_factories_and_types(
        self, empty_sources: DataSourceManager
    ):
        assert len(empty_sources.factories) == 0
        assert len(empty_sources.type_lookup) == 0

        class MyTestDatasource(Datasource):
            asset_types: ClassVar[List[Type[DataAsset]]] = []
            type: str = "my_test"

            @property
            @override
            def execution_engine_type(self) -> Type[ExecutionEngine]:
                return DummyExecutionEngine

        # One for each crud method: add, update, add_or_update, delete
        assert len(empty_sources.factories) == 4
        # type_lookup maps the MyTestDatasource.type str to the class and vis versa
        assert len(empty_sources.type_lookup) == 2
        assert "my_test" in empty_sources.type_lookup

    def test__new__registers_sources_factory_method(
        self, context_sources_cleanup: DataSourceManager
    ):
        expected_method_name = "add_my_test"

        ds_factory_method_initial = getattr(context_sources_cleanup, expected_method_name, None)
        assert ds_factory_method_initial is None, "Check test cleanup"

        class MyTestDatasource(Datasource):
            asset_types: ClassVar[List[Type[DataAsset]]] = []
            type: str = "my_test"

            @property
            @override
            def execution_engine_type(self) -> Type[ExecutionEngine]:
                return DummyExecutionEngine

        ds_factory_method_final = getattr(context_sources_cleanup, expected_method_name, None)

        assert ds_factory_method_final, (
            f"{MetaDatasource.__name__}.__new__ failed to add `{expected_method_name}()` method"
        )

    def test_registered_sources_factory_method_has_correct_signature(
        self, context_sources_cleanup: DataSourceManager
    ):
        expected_method_name = "add_my_test"

        class MyTestDatasource(Datasource):
            asset_types: ClassVar[List[Type[DataAsset]]] = []
            type: str = "my_test"
            extra_field: str

            @property
            @override
            def execution_engine_type(self) -> Type[ExecutionEngine]:
                return DummyExecutionEngine

        ds_factory_method = getattr(context_sources_cleanup, expected_method_name)

        ds_factory_method_sig = inspect.signature(ds_factory_method)
        my_ds_init_method_sig = inspect.signature(MyTestDatasource)
        print(f"{MyTestDatasource.__name__}.__init__()\n  {my_ds_init_method_sig}\n")
        print(f"{expected_method_name}()\n  {ds_factory_method_sig}\n")

        for i, param_name in enumerate(my_ds_init_method_sig.parameters):
            print(f"{i} {param_name} ", end="")

            if param_name in ["type", "assets"]:
                assert param_name not in ds_factory_method_sig.parameters, (
                    f"{param_name} should not be part of the `add_<DATASOURCE_TYPE>` method"
                )
                print("⏩")
                continue

            assert param_name in ds_factory_method_sig.parameters
            print("✅")

    def test__new__updates_asset_type_lookup(self, context_sources_cleanup: DataSourceManager):
        class FooAsset(DummyDataAsset):
            type: str = "foo"

        class BarAsset(DummyDataAsset):
            type: str = "bar"

        class FooBarDatasource(Datasource):
            asset_types: ClassVar[List[Type[DataAsset]]] = [FooAsset, BarAsset]
            type: str = "foo_bar"

            @property
            @override
            def execution_engine_type(self) -> Type[ExecutionEngine]:
                return DummyExecutionEngine

        print(f" type_lookup ->\n{pf(FooBarDatasource._type_lookup)}\n")
        asset_types = FooBarDatasource.asset_types
        assert asset_types, "No asset types have been declared"

        registered_type_names = [FooBarDatasource._type_lookup.get(t) for t in asset_types]
        for type_, name in zip(asset_types, registered_type_names):
            print(f"`{type_.__name__}` registered as '{name}'")
            assert name, f"{type.__name__} could not be retrieved"

        assert len(asset_types) == len(registered_type_names)


@pytest.mark.unit
class TestMisconfiguredMetaDatasource:
    def test_ds_type_field_not_set(self, empty_sources: DataSourceManager):
        with pytest.raises(
            TypeRegistrationError,
            match=r"`MissingTypeDatasource` is missing a `type` attribute",
        ):

            class MissingTypeDatasource(Datasource):
                @property
                @override
                def execution_engine_type(self) -> Type[ExecutionEngine]:
                    return DummyExecutionEngine

                @override
                def test_connection(self) -> None: ...  # type: ignore[override] # FIXME CoP

        # check that no types were registered
        assert len(empty_sources.type_lookup) < 1

    def test_ds_execution_engine_type_not_defined(self, empty_sources: DataSourceManager):
        class MissingExecEngineTypeDatasource(Datasource):
            type: str = "valid"

            @override
            def test_connection(self) -> None: ...  # type: ignore[override] # FIXME CoP

        with pytest.raises(NotImplementedError):
            MissingExecEngineTypeDatasource(name="name").get_execution_engine()

    def test_ds_assets_type_field_not_set(self, empty_sources: DataSourceManager):
        with pytest.raises(
            TypeRegistrationError,
            match="No `type` field found for `BadAssetDatasource.asset_types` -> `MissingTypeAsset` unable to register asset type",  # noqa: E501 # FIXME CoP
        ):

            class MissingTypeAsset(DataAsset):
                @override
                def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
                    raise NotImplementedError

                @override
                def get_batch(self, batch_request: BatchRequest) -> Batch:
                    raise NotImplementedError

            class BadAssetDatasource(Datasource):
                type: str = "valid"
                asset_types: ClassVar[List[Type[DataAsset]]] = [MissingTypeAsset]

                @property
                @override
                def execution_engine_type(self) -> Type[ExecutionEngine]:
                    return DummyExecutionEngine

                @override
                def test_connection(self) -> None: ...  # type: ignore[override] # FIXME CoP

        # check that no types were registered
        assert len(empty_sources.type_lookup) < 1

    def test_ds_test_connection_not_defined(self, empty_sources: DataSourceManager):
        class MissingTestConnectionDatasource(Datasource):
            type: str = "valid"

            @property
            @override
            def execution_engine_type(self) -> Type[ExecutionEngine]:
                return DummyExecutionEngine

        with pytest.raises(NotImplementedError):
            MissingTestConnectionDatasource(name="name").test_connection()


@pytest.mark.big
def test_minimal_ds_to_asset_flow(context_sources_cleanup):
    # 1. Define Datasource & Assets

    class SampleAsset(DataAsset):
        @override
        def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
            raise NotImplementedError

        @override
        def get_batch(self, batch_request: BatchRequest) -> Batch:
            raise NotImplementedError

    class RedAsset(SampleAsset):
        type = "red"

        def test_connection(self): ...  # type: ignore[explicit-override] # FIXME

    class BlueAsset(SampleAsset):
        type = "blue"

        @override
        def test_connection(self): ...

    class PurpleDatasource(Datasource):
        asset_types = [RedAsset, BlueAsset]
        type: str = "purple"

        @property
        @override
        def execution_engine_type(self) -> Type[ExecutionEngine]:
            return DummyExecutionEngine

        def test_connection(self): ...  # type: ignore[explicit-override] # FIXME

        def add_red_asset(self, asset_name: str) -> RedAsset:
            asset = RedAsset(name=asset_name)  # type: ignore[call-arg] # ?
            self._add_asset(asset=asset)
            return asset

    # 2. Get context
    context = get_context()

    # 3. Add a datasource
    purple_ds: Datasource = context.data_sources.add_purple("my_ds_name")

    # 4. Add a DataAsset
    red_asset: DataAsset = purple_ds.add_red_asset("my_asset_name")
    assert isinstance(red_asset, RedAsset)

    # 5. Get an asset by name - (method defined in parent `Datasource`)
    assert red_asset is purple_ds.get_asset("my_asset_name")


# Testing crud methods
DEFAULT_CRUD_DATASOURCE_NAME = "pandas_datasource"


@pytest.fixture
def context_config_data(
    file_dc_config_dir_init: pathlib.Path,
) -> Tuple[AbstractDataContext, pathlib.Path, pathlib.Path]:
    config_file_path = file_dc_config_dir_init / FileDataContext.GX_YML
    assert config_file_path.exists() is True
    context = get_gx_context(context_root_dir=file_dc_config_dir_init)
    data_dir = file_dc_config_dir_init.parent / "data"
    data_dir.mkdir()
    return context, config_file_path, data_dir


def assert_fluent_datasource_content(
    config_file_path: pathlib.Path, fluent_datasource_config: dict
):
    config = yaml.load(config_file_path.read_text())
    assert _FLUENT_DATASOURCES_KEY in config
    config_from_gx_yaml = config[_FLUENT_DATASOURCES_KEY]
    assert isinstance(config_from_gx_yaml, dict)
    config_from_gx_yaml_without_ids = _remove_ids(config_from_gx_yaml)
    assert config_from_gx_yaml_without_ids == fluent_datasource_config


def _remove_ids(config: dict) -> dict:
    for data_source in config.values():
        data_source.pop("id")
        for asset in data_source.get("assets", {}).values():
            asset.pop("id")
            for batch_definition in asset.get("batch_definitions", []):
                batch_definition.pop("id")

    return config


@pytest.fixture
def context_with_fluent_datasource(
    context_config_data: Tuple[AbstractDataContext, pathlib.Path, pathlib.Path],
) -> Tuple[AbstractDataContext, pathlib.Path, pathlib.Path]:
    context, config_file_path, data_dir = context_config_data
    assert len(context.data_sources.all()) == 0
    context.data_sources.add_pandas_filesystem(
        name=DEFAULT_CRUD_DATASOURCE_NAME,
        base_directory=data_dir,
        data_context_root_directory=config_file_path.parent,
    )
    assert len(context.data_sources.all()) == 1
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            DEFAULT_CRUD_DATASOURCE_NAME: {
                "base_directory": str(data_dir),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
        },
    )
    return context, config_file_path, data_dir


@pytest.mark.unit
def test_add_datasource(context_with_fluent_datasource):
    pass


@pytest.mark.unit
@pytest.mark.parametrize("use_positional_arg", [True, False])
def test_add_datasource_with_datasource_object(context_with_fluent_datasource, use_positional_arg):
    context, config_file_path, data_dir = context_with_fluent_datasource
    new_datasource = copy.deepcopy(context.data_sources.get(DEFAULT_CRUD_DATASOURCE_NAME))
    new_datasource.name = "new_datasource"
    if use_positional_arg:
        context.data_sources.add_pandas_filesystem(new_datasource)
    else:
        context.data_sources.add_pandas_filesystem(datasource=new_datasource)
    assert len(context.data_sources.all()) == 2
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            "pandas_datasource": {
                "base_directory": str(data_dir),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
            "new_datasource": {
                "base_directory": str(data_dir),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
        },
    )


@pytest.mark.unit
@pytest.mark.parametrize("use_positional_arg", [True, False])
def test_update_datasource(context_with_fluent_datasource, use_positional_arg):
    context, config_file_path, data_dir = context_with_fluent_datasource
    data_dir_2 = data_dir.parent / "data2"
    data_dir_2.mkdir()
    if use_positional_arg:
        context.data_sources.update_pandas_filesystem(
            DEFAULT_CRUD_DATASOURCE_NAME,
            base_directory=data_dir_2,
            data_context_root_directory=config_file_path.parent,
        )
    else:
        context.data_sources.update_pandas_filesystem(
            name=DEFAULT_CRUD_DATASOURCE_NAME,
            base_directory=data_dir_2,
            data_context_root_directory=config_file_path.parent,
        )
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            DEFAULT_CRUD_DATASOURCE_NAME: {
                "base_directory": str(data_dir_2),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
        },
    )


@pytest.mark.unit
@pytest.mark.parametrize("use_positional_arg", [True, False])
def test_update_datasource_with_datasource_object(
    context_with_fluent_datasource, use_positional_arg
):
    context, config_file_path, data_dir = context_with_fluent_datasource
    datasource = context.data_sources.get(DEFAULT_CRUD_DATASOURCE_NAME)
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            DEFAULT_CRUD_DATASOURCE_NAME: {
                "base_directory": str(data_dir),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
        },
    )

    # Add an asset and update datasource
    (data_dir / "1.csv").touch()
    if use_positional_arg:
        datasource.add_csv_asset("csv_asset")
    else:
        datasource.add_csv_asset(name="csv_asset")

    context.data_sources.update_pandas_filesystem(datasource=datasource)
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            DEFAULT_CRUD_DATASOURCE_NAME: {
                "base_directory": str(data_dir),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
                "assets": {
                    "csv_asset": {
                        "type": "csv",
                    },
                },
            },
        },
    )


@pytest.mark.unit
@pytest.mark.parametrize("use_positional_arg", [True, False])
def test_add_or_update_datasource_using_add(context_with_fluent_datasource, use_positional_arg):
    context, config_file_path, data_dir = context_with_fluent_datasource
    data_dir_2 = data_dir.parent / "data2"
    data_dir_2.mkdir()
    if use_positional_arg:
        context.data_sources.add_or_update_pandas_filesystem(
            f"{DEFAULT_CRUD_DATASOURCE_NAME}_2",
            base_directory=data_dir_2,
            data_context_root_directory=config_file_path.parent,
        )
    else:
        context.data_sources.add_or_update_pandas_filesystem(
            name=f"{DEFAULT_CRUD_DATASOURCE_NAME}_2",
            base_directory=data_dir_2,
            data_context_root_directory=config_file_path.parent,
        )
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            f"{DEFAULT_CRUD_DATASOURCE_NAME}_2": {
                "base_directory": str(data_dir_2),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
            DEFAULT_CRUD_DATASOURCE_NAME: {
                "base_directory": str(data_dir),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
        },
    )


@pytest.mark.unit
@pytest.mark.parametrize("use_positional_arg", [True, False])
def test_add_or_update_datasource_using_update(context_with_fluent_datasource, use_positional_arg):
    context, config_file_path, data_dir = context_with_fluent_datasource
    data_dir_2 = data_dir.parent / "data2"
    data_dir_2.mkdir()
    if use_positional_arg:
        context.data_sources.add_or_update_pandas_filesystem(
            DEFAULT_CRUD_DATASOURCE_NAME,
            base_directory=data_dir_2,
            data_context_root_directory=config_file_path.parent,
        )
    else:
        context.data_sources.add_or_update_pandas_filesystem(
            name=DEFAULT_CRUD_DATASOURCE_NAME,
            base_directory=data_dir_2,
            data_context_root_directory=config_file_path.parent,
        )
    assert_fluent_datasource_content(
        config_file_path=config_file_path,
        fluent_datasource_config={
            DEFAULT_CRUD_DATASOURCE_NAME: {
                "base_directory": str(data_dir_2),
                "data_context_root_directory": str(config_file_path.parent),
                "type": "pandas_filesystem",
            },
        },
    )


@pytest.mark.xfail(
    reason="There is a bug in context._save_config where deletions don't get persisted",
    run=True,
    strict=True,
)
@pytest.mark.unit
def test_delete_datasource(context_with_fluent_datasource):
    context, config_file_path, _data_dir = context_with_fluent_datasource
    context.data_sources.delete(name=DEFAULT_CRUD_DATASOURCE_NAME)
    assert_fluent_datasource_content(config_file_path, {})


@pytest.mark.unit
def test_legacy_delete_datasource_raises_deprecation_warning(
    context_with_fluent_datasource,
):
    context, _, _ = context_with_fluent_datasource
    with pytest.deprecated_call():
        context.data_sources.delete_pandas_filesystem(name=DEFAULT_CRUD_DATASOURCE_NAME)


if __name__ == "__main__":
    pytest.main([__file__, "-vv", "--log-level=DEBUG"])



import logging

import pytest
from pytest_mock import MockerFixture

from great_expectations.data_context import EphemeralDataContext
from great_expectations.datasource.fluent.redshift_datasource import (
    RedshiftConnectionDetails,
    RedshiftDsn,
    RedshiftSSLModes,
)

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def scheme():
    return "redshift+psycopg2"


@pytest.mark.unit
@pytest.mark.parametrize(
    "connection_input,expected_connection_string",
    [
        pytest.param(
            "redshift+psycopg2://user:password@host:1234/database?sslmode=allow",
            "redshift+psycopg2://user:password@host:1234/database?sslmode=allow",
            id="string type",
        ),
        pytest.param(
            {
                "user": "user",
                "password": "password",
                "host": "host",
                "port": 1234,
                "database": "database",
                "sslmode": RedshiftSSLModes.ALLOW,
            },
            "redshift+psycopg2://user:password@host:1234/database?sslmode=allow",
            id="dict type",
        ),
        pytest.param(
            RedshiftConnectionDetails(
                user="user",
                password="password",
                host="host",
                port=1234,
                database="database",
                sslmode=RedshiftSSLModes.ALLOW,
            ),
            "redshift+psycopg2://user:password@host:1234/database?sslmode=allow",
            id="dict type",
        ),
    ],
)
def test_create_engine_is_called_with_expected_kwargs(
    connection_input,
    expected_connection_string,
    sa,
    mocker: MockerFixture,
    ephemeral_context_with_defaults: EphemeralDataContext,
    scheme,
):
    create_engine_spy = mocker.patch.object(sa, "create_engine")

    context = ephemeral_context_with_defaults
    data_source = context.data_sources.add_redshift(
        name="redshift_test",
        connection_string=connection_input,
    )
    data_source.get_engine()  # we will verify that the correct connection details are used when getting the engine  # noqa: E501

    expected_kwargs = RedshiftDsn(
        expected_connection_string,
        scheme=scheme,
    )

    create_engine_spy.assert_called_once_with(expected_kwargs)


@pytest.mark.unit
def test_value_error_raised_if_invalid_connection_detail_inputs(
    sa,
    ephemeral_context_with_defaults: EphemeralDataContext,
    scheme,
):
    user = "user"
    password = "password"
    host = "host"
    port = 1234
    database = "database"
    sslmode = "INVALID"

    with pytest.raises(ValueError):
        RedshiftConnectionDetails(
            user=user,
            password=password,
            host=host,
            port=port,
            database=database,
            sslmode=sslmode,  # type: ignore[arg-type] # Ignore this for purpose of the test
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, Generic, Mapping, Optional, Sequence, Union

import numpy as np
import pandas as pd
from typing_extensions import override

from great_expectations.compatibility.sqlalchemy import (
    Column,
    MetaData,
    Table,
    TextClause,
    TypeEngine,
    create_engine,
    insert,
    sqltypes,
)
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from tests.integration.sql_session_manager import (
    ConnectionDetails,
    SessionSQLEngineManager,
)
from tests.integration.test_utils.data_source_config.base import BatchTestSetup, _ConfigT

if TYPE_CHECKING:
    import sqlalchemy as sa

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _TableData:
    name: str
    df: pd.DataFrame
    table: Table


InferrableTypesLookup = dict[type[Any], Union[type[TypeEngine], TypeEngine]]

InferredColumnTypes = dict[str, Union[type[TypeEngine], TypeEngine]]


class SQLBatchTestSetup(BatchTestSetup[_ConfigT, TableAsset], ABC, Generic[_ConfigT]):
    SCHEMA_PREFIX = "test_"

    @property
    @abstractmethod
    def connection_string(self) -> str:
        """Connection string used to connect to SQL backend."""

    @property
    @abstractmethod
    def use_schema(self) -> bool:
        """Whether to use a schema when connecting to SQL backend.

        If `True`, a schema will be automatically created.
        """

    @property
    def inferrable_types_lookup(self) -> InferrableTypesLookup:
        """Dict of Python type keys mapped to SQL dialect-specific SqlAlchemy types."""
        # implementations of the class can override this if more specific types are required
        return {
            str: sqltypes.VARCHAR,
            int: sqltypes.INTEGER,
            float: sqltypes.DECIMAL,
            bool: sqltypes.BOOLEAN,
            date: sqltypes.DATE,
            datetime: sqltypes.DATETIME,
            pd.Timestamp: sqltypes.DATETIME,
        }

    def __init__(
        self,
        config: _ConfigT,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        table_name: Optional[str] = None,  # Overrides random table name generation
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> None:
        self.engine_manager = engine_manager
        self.extra_data = extra_data
        self.metadata = MetaData()
        self._user_specified_table_name = table_name
        super().__init__(config, data, context=context)

    @override
    def make_batch(self) -> Batch:
        return (
            self.make_asset()
            .add_batch_definition_whole_table(name=self._random_resource_name())
            .get_batch()
        )

    @cached_property
    def table_name(self) -> str:
        return self.main_table_data.name

    @cached_property
    def main_table_data(self) -> _TableData:
        name = self._user_specified_table_name or self._create_table_name()
        return self._create_table_data(
            name=name,
            df=self.data,
            column_types=self.config.column_types or {},
        )

    @cached_property
    def extra_table_data(self) -> Mapping[str, _TableData]:
        return {
            label: self._create_table_data(
                name=self._create_table_name(label),
                df=df,
                column_types=self.config.extra_column_types.get(label, {}),
            )
            for label, df in self.extra_data.items()
        }

    @cached_property
    def tables(self) -> Sequence[Table]:
        extra_tables = [td.table for td in self.extra_table_data.values()]
        return [self.main_table_data.table, *extra_tables]

    @cached_property
    def schema(self) -> Union[str, None]:
        if self.use_schema:
            return f"{self.SCHEMA_PREFIX}{self._random_resource_name()}"
        else:
            return None

    def _get_engine(self) -> tuple[sa.engine.Engine, Callable[[], None]]:
        if self.engine_manager:
            connection_details = ConnectionDetails(
                connection_string=self.connection_string,
            )
            engine = self.engine_manager.get_engine(connection_details)
            return engine, lambda: None
        else:
            engine = create_engine(url=self.connection_string)
            return engine, engine.dispose

    @override
    def setup(self) -> None:
        engine, cleanup = self._get_engine()

        with engine.connect() as conn, conn.begin():
            # create schema if needed

            if self.schema:
                logger.info(f"CREATING SCHEMA {self.schema}")
                conn.execute(TextClause(f"CREATE SCHEMA {self.schema}"))

            # create tables
            all_table_data = self._ensure_all_table_data_created()
            self.metadata.create_all(engine)

            # insert data
            for table_data in all_table_data:
                # pd.DataFrame(...).to_dict("index") returns a dictionary where the keys are the row
                # index and the values are a dict of column names mapped to column values.
                # Then we pass that list of dicts in as parameters to our insert statement.
                #   INSERT INTO test_table (my_int_column, my_str_column) VALUES (?, ?)
                #   [...] [('1', 'foo'), ('2', 'bar')]
                df = table_data.df.replace(np.nan, None)
                values = list(df.to_dict("index").values())
                conn.execute(insert(table_data.table), values)
        cleanup()

    @override
    def teardown(self) -> None:
        engine, cleanup = self._get_engine()
        for table in self.tables:
            table.drop(engine)
        if self.schema:
            with engine.connect() as conn, conn.begin():
                logger.info(f"DROPPING SCHEMA {self.schema}")
                conn.execute(TextClause(f"DROP SCHEMA {self.schema}"))
        cleanup()

    def _create_table_name(self, label: Optional[str] = None) -> str:
        parts = ["expectation_test_table", label, self._random_resource_name()]
        return "_".join([part for part in parts if part])

    def _ensure_all_table_data_created(self) -> Sequence[_TableData]:
        return [self.main_table_data, *self.extra_table_data.values()]

    def _create_table_data(
        self, name: str, df: pd.DataFrame, column_types: Mapping[str, type[TypeEngine]]
    ) -> _TableData:
        columns = self._get_column_types(df=df, column_types=column_types)
        table = self._create_table(name, columns=columns)
        return _TableData(
            name=name,
            df=df,
            table=table,
        )

    def _create_table(self, name: str, columns: InferredColumnTypes) -> Table:
        column_list = [Column(col_name, col_type) for col_name, col_type in columns.items()]
        return Table(name, self.metadata, *column_list, schema=self.schema)

    def _get_column_types(
        self,
        df: pd.DataFrame,
        column_types: Mapping[str, type[TypeEngine]],
    ) -> InferredColumnTypes:
        all_column_types = self._infer_column_types(df)
        # prefer explicit types if they're provided
        all_column_types.update(column_types)
        untyped_columns = set(df.columns) - set(all_column_types.keys())
        if untyped_columns:
            config_class_name = self.config.__class__.__name__
            message = (
                f"Unable to infer types for the following column(s): "
                f"{', '.join(untyped_columns)}. \n"
                f"Please provide the missing types as the `column_types` "
                f"parameter when \ninstantiating {config_class_name}."
            )
            raise RuntimeError(message)
        return all_column_types

    def _infer_column_types(self, data: pd.DataFrame) -> InferredColumnTypes:
        inferred_column_types: InferredColumnTypes = {}
        for column, value_list in data.to_dict("list").items():
            non_null_value_list = [val for val in value_list if val is not None]
            if not non_null_value_list:
                # if we have an all null column, just arbitrarily use INTEGER
                inferred_column_types[str(column)] = sqltypes.INTEGER
            else:
                python_type = type(non_null_value_list[0])
                if not all(isinstance(val, python_type) for val in non_null_value_list):
                    raise RuntimeError(
                        f"Cannot infer type of column {column}. "
                        "Please provide an explicit column type in the test config."
                    )
                inferred_type = self.inferrable_types_lookup.get(python_type)
                if inferred_type:
                    inferred_column_types[str(column)] = inferred_type
        return inferred_column_types



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Sequence, Union

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import (
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.query_metric_provider import (
    QueryMetricProvider,
)

if TYPE_CHECKING:
    from great_expectations.compatibility import pyspark


class QueryRowCount(QueryMetricProvider):
    metric_name = "query.row_count"
    value_keys = ("query",)

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ) -> int:
        batch_selectable, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        query = cls._get_query_from_metric_value_kwargs(metric_value_kwargs)
        substituted_batch_subquery = (
            cls._get_substituted_batch_subquery_from_query_and_batch_selectable(
                query=query,
                batch_selectable=batch_selectable,
                execution_engine=execution_engine,
            )
        )
        count_column_name = "unexpected_row_count"
        row_count_query = (
            f"SELECT COUNT(*) as {count_column_name} FROM "
            f"({substituted_batch_subquery}) AS substituted_batch_subquery"
        )
        result: Union[Sequence[sa.Row[Any]], Any] = execution_engine.execute_query(
            sa.text(row_count_query)
        ).fetchone()
        return int(result[0])

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ) -> int:
        query = cls._get_query_from_metric_value_kwargs(metric_value_kwargs)

        df: pyspark.DataFrame
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )

        df.createOrReplaceTempView("tmp_view")
        query = query.format(batch="tmp_view")

        engine: pyspark.SparkSession = execution_engine.spark
        return engine.sql(query).count()



from __future__ import annotations

import json
import logging
from typing import ClassVar

import requests
from packaging import version
from typing_extensions import TypedDict

logger = logging.getLogger(__name__)


class _PyPIPackageInfo(TypedDict):
    version: str


class _PyPIPackageData(TypedDict):
    info: _PyPIPackageInfo


class _VersionChecker:
    _LATEST_GX_VERSION_CACHE: ClassVar[version.Version | None] = None

    _BASE_PYPI_URL: ClassVar[str] = "https://pypi.org/pypi"
    _PYPI_GX_ENDPOINT: ClassVar[str] = f"{_BASE_PYPI_URL}/great_expectations/json"

    def __init__(self, user_version: str) -> None:
        self._user_version = version.Version(user_version)

    def check_if_using_latest_gx(self) -> bool:
        pypi_version: version.Version | None
        if self._LATEST_GX_VERSION_CACHE:
            pypi_version = self._LATEST_GX_VERSION_CACHE
        else:
            pypi_version = self._get_latest_version_from_pypi()
            if not pypi_version:
                logger.debug("Could not compare with latest PyPI version; skipping check.")
                return True

        if self._is_using_outdated_release(pypi_version):
            self._warn_user(pypi_version)
            return False
        return True

    def _get_latest_version_from_pypi(self) -> version.Version | None:
        response_json: _PyPIPackageData | None = None
        try:
            response = requests.get(self._PYPI_GX_ENDPOINT)
            response.raise_for_status()
            response_json = response.json()
        except json.JSONDecodeError as jsonError:
            logger.debug(f"Failed to parse PyPI API response into JSON: {jsonError}")
        except requests.HTTPError as http_err:
            logger.debug(f"An HTTP error occurred when trying to hit PyPI API: {http_err}")
        except requests.Timeout as timeout_exc:
            logger.debug(f"Failed to hit the PyPI API due to a timeout error: {timeout_exc}")
        except requests.ConnectionError as connection_err:
            logger.debug(f"Failed to hit the PyPI API due to a connection error: {connection_err}")

        if not response_json:
            return None

        # Structure should be guaranteed but let's be defensive in case PyPI changes.
        info = response_json.get("info", {})
        pkg_version = info.get("version")
        if not pkg_version:
            logger.debug("Successfully hit PyPI API but payload structure is not as expected.")
            return None

        pypi_version = version.Version(pkg_version)
        # update the _LATEST_GX_VERSION_CACHE
        self.__class__._LATEST_GX_VERSION_CACHE = pypi_version
        return pypi_version

    def _is_using_outdated_release(self, pypi_version: version.Version) -> bool:
        return pypi_version > self._user_version

    def _warn_user(self, pypi_version: version.Version) -> None:
        logger.warning(
            f"You are using great_expectations version {self._user_version}; "
            f"however, version {pypi_version} is available.\nYou should consider "
            "upgrading via `pip install great_expectations --upgrade`.\n"
        )



from .exceptions import (
    BatchDefinitionError,
    BatchDefinitionNotFoundError,
    BatchFilterError,
    BatchSpecError,
    CheckpointError,
    CheckpointNotFoundError,
    CheckpointRunWithoutValidationDefinitionError,
    ClassInstantiationError,
    ConfigNotFoundError,
    DataAssetNotFoundError,
    DatabaseConnectionError,
    DataContextError,
    DataContextRequiredError,
    DatasourceError,
    DatasourceInitializationError,
    DatasourceKeyPairAuthBadPassphraseError,
    DatasourceNotFoundError,
    ExecutionEngineError,
    ExpectationNotFoundError,
    ExpectationSuiteError,
    ExpectationSuiteNotFoundError,
    GitIgnoreScaffoldingError,
    GreatExpectationsError,
    GreatExpectationsTypeError,
    GreatExpectationsValidationError,
    GXCloudConfigurationError,
    GXCloudError,
    InvalidBaseYamlConfigError,
    InvalidBatchIdError,
    InvalidBatchRequestError,
    InvalidBatchSpecError,
    InvalidCacheValueError,
    InvalidConfigError,
    InvalidConfigurationYamlError,
    InvalidDataContextConfigError,
    InvalidDataContextKeyError,
    InvalidExpectationConfigurationError,
    InvalidExpectationKwargsError,
    InvalidKeyError,
    InvalidMetricAccessorDomainKwargsKeyError,
    InvalidTopLevelConfigKeyError,
    MetricComputationError,
    MetricError,
    MetricProviderError,
    MetricResolutionError,
    MissingConfigVariableError,
    MissingDataContextError,
    NoDataDocsError,
    PluginClassNotFoundError,
    PluginModuleNotFoundError,
    RenderingError,
    SamplerError,
    SorterError,
    StoreBackendError,
    StoreBackendTransientError,
    StoreBackendUnsupportedResourceTypeError,
    StoreConfigurationError,
    StoreError,
    SuiteParameterError,
    UnavailableMetricError,
    UnhashableColumnError,
    UnsupportedConfigVersionError,
    ValidationError,
)
from .resource_freshness import (
    BatchDefinitionNotAddedError,
    BatchDefinitionNotFreshError,
    CheckpointNotAddedError,
    CheckpointNotFreshError,
    CheckpointRelatedResourcesFreshnessError,
    ExpectationSuiteNotAddedError,
    ExpectationSuiteNotFreshError,
    ResourceFreshnessAggregateError,
    ResourceFreshnessError,
    ValidationDefinitionNotAddedError,
    ValidationDefinitionNotFreshError,
    ValidationDefinitionRelatedResourcesFreshnessError,
)



import pytest

from great_expectations.render.renderer import EmailRenderer


@pytest.mark.big
def test_EmailRenderer_get_report_element():
    email_renderer = EmailRenderer()

    # these should all be caught
    assert email_renderer._get_report_element(docs_link=None) is None
    assert email_renderer._get_report_element(docs_link=1) is None
    assert email_renderer._get_report_element(docs_link=email_renderer) is None

    # this should work
    assert email_renderer._get_report_element(docs_link="i_should_work") is not None


@pytest.mark.unit
def test_EmailRenderer_render(v1_checkpoint_result):
    email_renderer = EmailRenderer()
    _, raw_html = email_renderer.render(checkpoint_result=v1_checkpoint_result)
    html_blocks = raw_html.split("\n")

    assert html_blocks == [
        "<p><strong><h3><u>my_bad_suite</u></h3></strong></p>",
        "<p><strong>Batch Validation Status</strong>: Failed ❌</p>",
        "<p><strong>Expectation Suite Name</strong>: my_bad_suite</p>",
        "<p><strong>Data Asset Name</strong>: my_first_asset</p>",
        "<p><strong>Run ID</strong>: __no_run_id__</p>",
        "<p><strong>Batch ID</strong>: my_batch</p>",
        "<p><strong>Summary</strong>: <strong>3</strong> of <strong>5</strong> expectations were met</p><br><p><strong><h3><u>my_good_suite</u></h3></strong></p>",  # noqa: E501 # FIXME CoP
        "<p><strong>Batch Validation Status</strong>: Success 🎉</p>",
        "<p><strong>Expectation Suite Name</strong>: my_good_suite</p>",
        "<p><strong>Data Asset Name</strong>: __no_asset_name__</p>",
        "<p><strong>Run ID</strong>: my_run_id</p>",
        "<p><strong>Batch ID</strong>: my_other_batch</p>",
        "<p><strong>Summary</strong>: <strong>1</strong> of <strong>1</strong> expectations were met</p>",  # noqa: E501 # FIXME CoP
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.core.types import Comparable  # noqa: TC001 # FIXME CoP
from great_expectations.expectations.expectation import (
    ColumnAggregateExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import COLUMN_DESCRIPTION
from great_expectations.render import (
    LegacyDescriptiveRendererType,
    LegacyRendererType,
    RenderedStringTemplateContent,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    handle_strict_min_max,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the column mean to be between a minimum value and a maximum value (inclusive)."
)
MIN_VALUE_DESCRIPTION = "The minimum value for the column mean."
MAX_VALUE_DESCRIPTION = "The maximum value for the column mean."
STRICT_MIN_DESCRIPTION = (
    "If True, the column mean must be strictly larger than min_value, default=False"
)
STRICT_MAX_DESCRIPTION = (
    "If True, the column mean must be strictly smaller than max_value, default=False"
)
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.NUMERIC.value]


class ExpectColumnMeanToBeBetween(ColumnAggregateExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnMeanToBeBetween is a \
    Column Aggregate Expectation.

    Column Aggregate Expectations are one of the most common types of Expectation.
    They are evaluated for a single column, and produce an aggregate Metric, such as a mean, standard deviation, number of unique values, column type, etc.
    If that Metric meets the conditions you set, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        min_value (float or None): \
            {MIN_VALUE_DESCRIPTION}
        max_value (float or None): \
            {MAX_VALUE_DESCRIPTION}
        strict_min (boolean): \
            {STRICT_MIN_DESCRIPTION}
        strict_max (boolean): \
            {STRICT_MAX_DESCRIPTION}

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Notes:
        * min_value and max_value are both inclusive unless strict_min or strict_max are set to True.
        * If min_value is None, then max_value is treated as an upper bound.
        * If max_value is None, then min_value is treated as a lower bound.
        * observed_value field in the result object is customized for this expectation to be a float \
            representing the true mean for the column

    See Also:
        [ExpectColumnMedianToBeBetween](https://greatexpectations.io/expectations/expect_column_median_to_be_between)
        [ExpectColumnStdevToBeBetween](https://greatexpectations.io/expectations/expect_column_stdev_to_be_between)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1       1
            1 	1.3     7
            2 	.8      2.5
            3   2       3

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnMeanToBeBetween(
                    column="test",
                    min_value=1,
                    max_value=3
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 1.275
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnMeanToBeBetween(
                    column="test2",
                    min_value=1,
                    max_value=3
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 3.375
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    min_value: Optional[Comparable] = pydantic.Field(
        default=None, description=MIN_VALUE_DESCRIPTION
    )
    max_value: Optional[Comparable] = pydantic.Field(
        default=None, description=MAX_VALUE_DESCRIPTION
    )
    strict_min: Union[bool, SuiteParameterDict] = pydantic.Field(
        default=False, description=STRICT_MIN_DESCRIPTION
    )
    strict_max: Union[bool, SuiteParameterDict] = pydantic.Field(
        default=False, description=STRICT_MAX_DESCRIPTION
    )

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "column aggregate expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    # Setting necessary computation metric dependencies and defining kwargs, as well as assigning kwargs default values\  # noqa: E501 # FIXME CoP
    metric_dependencies = ("column.mean",)
    success_keys = (
        "min_value",
        "strict_min",
        "max_value",
        "strict_max",
    )

    args_keys = (
        "column",
        "min_value",
        "max_value",
        "strict_min",
        "strict_max",
    )

    class Config:
        title = "Expect column mean to be between"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[ExpectColumnMeanToBeBetween]) -> None:
            ColumnAggregateExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    @override
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("min_value", [RendererValueType.NUMBER, RendererValueType.DATETIME]),
            ("max_value", [RendererValueType.NUMBER, RendererValueType.DATETIME]),
            ("strict_min", RendererValueType.BOOLEAN),
            ("strict_max", RendererValueType.BOOLEAN),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.min_value and not params.max_value:
            template_str = "mean may have any numerical value."
        else:
            at_least_str = "greater than or equal to"
            if params.strict_min:
                at_least_str = cls._get_strict_min_string(
                    renderer_configuration=renderer_configuration
                )
            at_most_str = "less than or equal to"
            if params.strict_max:
                at_most_str = cls._get_strict_max_string(
                    renderer_configuration=renderer_configuration
                )

            if params.min_value and params.max_value:
                template_str = (
                    f"mean must be {at_least_str} $min_value and {at_most_str} $max_value."
                )
            elif not params.min_value:
                template_str = f"mean must be {at_most_str} $max_value."
            else:
                template_str = f"mean must be {at_least_str} $min_value."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @override
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(  # type: ignore[override] # TODO: Fix this type ignore
        cls,
        configuration: ExpectationConfiguration,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            [
                "column",
                "min_value",
                "max_value",
                "row_condition",
                "condition_parser",
                "strict_min",
                "strict_max",
            ],
        )

        template_str = ""
        if (params["min_value"] is None) and (params["max_value"] is None):
            template_str = "mean may have any numerical value."
        else:
            at_least_str, at_most_str = handle_strict_min_max(params)

            if params["min_value"] is not None and params["max_value"] is not None:
                template_str = (
                    f"mean must be {at_least_str} $min_value and {at_most_str} $max_value."
                )
            elif params["min_value"] is None:
                template_str = f"mean must be {at_most_str} $max_value."
            elif params["max_value"] is None:
                template_str = f"mean must be {at_least_str} $min_value."

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    @classmethod
    @renderer(renderer_type=LegacyDescriptiveRendererType.STATS_TABLE_MEAN_ROW)
    def _descriptive_stats_table_mean_row_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        assert result, "Must pass in result."
        return [
            {
                "content_block_type": "string_template",
                "string_template": {
                    "template": "Mean",
                    "tooltip": {"content": "expect_column_mean_to_be_between"},
                },
            },
            f"{result.result['observed_value']:.2f}",
        ]

    @override
    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        return self._validate_metric_value_between(
            metric_name="column.mean",
            metrics=metrics,
            runtime_configuration=runtime_configuration,
            execution_engine=execution_engine,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from great_expectations.compatibility.typing_extensions import override
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.metrics.column.descriptive_stats import DescriptiveStats
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


class ColumnDescriptiveStats(ColumnAggregateMetricProvider):
    metric_name = "column.descriptive_stats"

    @metric_value(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        metrics: dict[str, int],
        **kwargs,
    ) -> DescriptiveStats:
        return DescriptiveStats(
            min=metrics["column.min"],
            max=metrics["column.max"],
            mean=metrics["column.mean"],
            standard_deviation=metrics["column.standard_deviation"],
        )

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        metrics: dict[str, int],
        **kwargs,
    ) -> DescriptiveStats:
        return DescriptiveStats(
            min=metrics["column.min"],
            max=metrics["column.max"],
            mean=metrics["column.mean"],
            standard_deviation=metrics["column.standard_deviation"],
        )

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        metrics: dict[str, int],
        **kwargs,
    ) -> DescriptiveStats:
        return DescriptiveStats(
            min=metrics["column.min"],
            max=metrics["column.max"],
            mean=metrics["column.mean"],
            standard_deviation=metrics["column.standard_deviation"],
        )

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        dependencies: dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )
        dependencies["column.min"] = MetricConfiguration(
            metric_name="column.min",
            metric_domain_kwargs=metric.metric_domain_kwargs,
            metric_value_kwargs=None,
        )
        dependencies["column.max"] = MetricConfiguration(
            metric_name="column.max",
            metric_domain_kwargs=metric.metric_domain_kwargs,
            metric_value_kwargs=None,
        )
        dependencies["column.mean"] = MetricConfiguration(
            metric_name="column.mean",
            metric_domain_kwargs=metric.metric_domain_kwargs,
            metric_value_kwargs=None,
        )
        dependencies["column.standard_deviation"] = MetricConfiguration(
            metric_name="column.standard_deviation",
            metric_domain_kwargs=metric.metric_domain_kwargs,
            metric_value_kwargs=None,
        )
        return dependencies



import uuid
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Union

import pytest

from great_expectations.compatibility.pydantic import (
    error_wrappers as pydantic_error_wrappers,
)
from great_expectations.core import (
    ExpectationValidationResult,
)
from great_expectations.expectations.expectation import Expectation
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)

if TYPE_CHECKING:
    from great_expectations.render.renderer_configuration import AddParamArgs


def mock_expectation_validation_result_from_expectation_configuration(
    expectation_configuration: ExpectationConfiguration,
    fake_result: dict,
) -> ExpectationValidationResult:
    return ExpectationValidationResult(
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        expectation_config=expectation_configuration,
        result=fake_result,
        success=True,
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "expectation_type,kwargs,runtime_configuration,fake_result,include_column_name",
    [
        (
            "expect_table_row_count_to_equal",
            {"value": 3},
            None,
            {"observed_value": 3},
            True,
        ),
        (
            "expect_table_row_count_to_equal",
            {"value": 3},
            {},
            {"observed_value": 3},
            True,
        ),
        (
            "expect_table_row_count_to_equal",
            {"value": 3},
            {"include_column_name": False},
            {"observed_value": 3},
            False,
        ),
    ],
)
def test_successful_renderer_configuration_instantiation(
    expectation_type: str,
    kwargs: dict,
    runtime_configuration: dict,
    fake_result: dict,
    include_column_name: bool,
):
    expectation_configuration = ExpectationConfiguration(
        type=expectation_type,
        kwargs=kwargs,
    )
    renderer_configuration = RendererConfiguration(
        configuration=expectation_configuration,
        runtime_configuration=runtime_configuration,
    )
    assert renderer_configuration.expectation_type == expectation_type
    assert renderer_configuration.kwargs == kwargs
    assert renderer_configuration.include_column_name is include_column_name

    expectation_validation_result = (
        mock_expectation_validation_result_from_expectation_configuration(
            expectation_configuration=expectation_configuration,
            fake_result=fake_result,
        )
    )
    renderer_configuration = RendererConfiguration(
        result=expectation_validation_result,
        runtime_configuration=runtime_configuration,
    )
    assert renderer_configuration.expectation_type == expectation_type
    assert renderer_configuration.kwargs == kwargs
    assert renderer_configuration.include_column_name is include_column_name


@pytest.mark.unit
@pytest.mark.parametrize(
    "expectation_type,kwargs, rendered_value",
    [
        (
            "expect_column_values_to_be_in_set",
            {
                "value_set": [1, 2, 3],
                "column": "Survived",
                "condition_parser": "great_expectations",
                "row_condition": 'PClass=="1st"',
            },
            ['PClass=="1st"'],
        ),
        (
            "expect_column_values_to_be_in_set",
            {
                "value_set": [1, 2, 3],
                "column": "Survived",
                "condition_parser": "great_expectations",
                "row_condition": 'col("foo") > 5',
            },
            ["foo", "> 5"],
        ),
        (
            "expect_column_values_to_be_in_set",
            {
                "value_set": [1, 2, 3],
                "column": "Survived",
                "condition_parser": "great_expectations",
                "row_condition": "foo == 1 | foo == 2",
            },
            ["foo == 1", "foo == 2"],
        ),
        (
            "expect_column_values_to_be_in_set",
            {
                "value_set": [1, 2, 3],
                "column": "Survived",
                "condition_parser": "great_expectations",
                "row_condition": "foo ~ 4",
            },
            ["foo", "4"],
        ),
        (
            "expect_column_values_to_be_in_set",
            {
                "value_set": [1, 2, 3],
                "column": "Survived",
                "condition_parser": "great_expectations",
                "row_condition": "foo > 1 & foo < 3",
            },
            ["foo > 1", "foo < 3"],
        ),
    ],
)
def test_successful_renderer_row_condition_params(
    expectation_type: str,
    kwargs: dict,
    rendered_value: list[str],
):
    expectation_configuration = ExpectationConfiguration(
        type=expectation_type,
        kwargs=kwargs,
    )
    renderer_configuration = RendererConfiguration(
        configuration=expectation_configuration,
        runtime_configuration={},
    )

    for index, value in enumerate(rendered_value):
        assert getattr(renderer_configuration.params, f"row_condition__{index}").value == value


@pytest.mark.unit
@pytest.mark.xfail(
    reason="As of v0.15.46 test will fail until RendererConfiguration._validate_configuration_or_result is re-enabled.",  # noqa: E501 # FIXME CoP
    strict=True,
)
def test_failed_renderer_configuration_instantiation():
    with pytest.raises(pydantic_error_wrappers.ValidationError) as e:
        RendererConfiguration(
            runtime_configuration={},
        )
    assert any(
        str(error_wrapper_exc)
        == "RendererConfiguration must be passed either configuration or result."
        for error_wrapper_exc in [error_wrapper.exc for error_wrapper in e.value.raw_errors]
    )


class NotString:
    def __str__(self):
        raise TypeError("I'm not a string")


@pytest.mark.unit
@pytest.mark.parametrize(
    "param_type,value",
    [
        (RendererValueType.STRING, NotString()),
        (RendererValueType.NUMBER, "ABC"),
        (RendererValueType.BOOLEAN, 3),
        (RendererValueType.ARRAY, 3),
    ],
)
def test_renderer_configuration_add_param_validation(
    param_type: RendererValueType, value: Union[NotString, str, int]
):
    expectation_configuration = ExpectationConfiguration(
        type="expect_table_row_count_to_equal",
        kwargs={"value": value},
    )
    renderer_configuration = RendererConfiguration(configuration=expectation_configuration)
    with pytest.raises(pydantic_error_wrappers.ValidationError) as e:
        renderer_configuration.add_param(name="value", param_type=param_type)

    if param_type is RendererValueType.STRING:
        exception_message = (
            f"Value was unable to be represented as a {RendererValueType.STRING}: I'm not a string"
        )
    else:
        exception_message = f"Param type: <{param_type}> does not match value: <{value}>."
    assert any(
        str(error_wrapper_exc) == exception_message
        for error_wrapper_exc in [error_wrapper.exc for error_wrapper in e.value.raw_errors]
    )


@pytest.mark.unit
def test_add_param_args():
    expectation_configuration = ExpectationConfiguration(
        type="expect_column_value_z_scores_to_be_less_than",
        kwargs={"column": "foo", "threshold": 2, "double_sided": False, "mostly": 1.0},
    )
    renderer_configuration = RendererConfiguration(configuration=expectation_configuration)

    add_param_args: AddParamArgs = (
        ("column", RendererValueType.STRING),
        ("threshold", RendererValueType.NUMBER),
        ("double_sided", RendererValueType.BOOLEAN),
        ("mostly", RendererValueType.NUMBER),
    )
    for name, param_type in add_param_args:
        renderer_configuration.add_param(name=name, param_type=param_type)

    params = renderer_configuration.params

    assert params.column
    assert params.threshold
    assert params.double_sided
    assert params.mostly


@pytest.mark.unit
def test_template_str_setter():
    expectation_configuration = ExpectationConfiguration(
        type="expect_column_value_z_scores_to_be_less_than",
        kwargs={"column": "foo", "threshold": 2, "double_sided": False, "mostly": 1.0},
    )
    renderer_configuration = RendererConfiguration(configuration=expectation_configuration)

    template_str = "My rendered string"
    renderer_configuration.template_str = template_str

    assert renderer_configuration.template_str == template_str


@pytest.mark.unit
def test_add_array_params():
    expectation_configuration = ExpectationConfiguration(
        type="expect_column_values_to_match_like_pattern_list",
        kwargs={"column": "foo", "like_pattern_list": ["%", "_"], "match_on": "any", "mostly": 1.0},
    )
    renderer_configuration = RendererConfiguration(configuration=expectation_configuration)
    renderer_configuration.add_param(name="like_pattern_list", param_type=RendererValueType.ARRAY)

    array_param_name = "like_pattern_list"
    param_prefix = array_param_name + "_"

    renderer_configuration = Expectation._add_array_params(
        array_param_name=array_param_name,
        param_prefix=param_prefix,
        renderer_configuration=renderer_configuration,
    )
    params = renderer_configuration.params

    assert params.like_pattern_list_0
    assert params.like_pattern_list_1

    array_string = Expectation._get_array_string(
        array_param_name=array_param_name,
        param_prefix=param_prefix,
        renderer_configuration=renderer_configuration,
    )

    assert array_string == "$like_pattern_list_0 $like_pattern_list_1"


@pytest.mark.parametrize(
    ("value", "type"),
    [
        ([1, 2, 3], RendererValueType.ARRAY),
        (True, RendererValueType.BOOLEAN),
        (datetime.now(tz=timezone.utc), RendererValueType.DATETIME),
        (3.14, RendererValueType.NUMBER),
        ({"foo": "bar"}, RendererValueType.OBJECT),
        ("hello world!", RendererValueType.STRING),
        (uuid.uuid4(), RendererValueType.STRING),
        (None, RendererValueType.STRING),
    ],
)
@pytest.mark.unit
def test_from_value(value: Any, type: RendererValueType) -> None:
    assert RendererValueType.from_value(value) == type



from __future__ import annotations

import copy
import logging
import pathlib
from typing import (
    TYPE_CHECKING,
    Generator,
    Literal,
    Optional,
    Tuple,
)
from unittest.mock import ANY

import pytest
from sqlalchemy.exc import SQLAlchemyError

import great_expectations.exceptions as ge_exceptions
from great_expectations.compatibility.pydantic import ValidationError
from great_expectations.core.batch_spec import SqlAlchemyDatasourceBatchSpec
from great_expectations.core.id_dict import IDDict
from great_expectations.core.partitioners import (
    ColumnPartitioner,
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
    PartitionerColumnValue,
    PartitionerDatetimePart,
    PartitionerDividedInteger,
    PartitionerModInteger,
    PartitionerMultiColumnValue,
)
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.datasource.fluent.batch_request import (
    BatchParameters,
    BatchRequest,
)
from great_expectations.datasource.fluent.interfaces import TestConnectionError
from great_expectations.datasource.fluent.postgres_datasource import (
    PostgresDatasource,
)
from great_expectations.datasource.fluent.sql_datasource import (
    SqlPartitionerYearAndMonth,
    TableAsset,
)
from great_expectations.exceptions.exceptions import NoAvailableBatchesError
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from tests.datasource.fluent.conftest import (
    _DEFAULT_TEST_MONTHS,
    _DEFAULT_TEST_YEARS,
    CreateSourceFixture,
)
from tests.sqlalchemy_test_doubles import Dialect, MockSaEngine, MockSaInspector

if TYPE_CHECKING:
    from unittest.mock import Mock  # noqa: TID251 # FIXME CoP

    from pytest_mock import MockFixture

    from great_expectations.datasource.fluent.interfaces import (
        BatchMetadata,
    )

# We set a default time range that we use for testing.

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def mock_test_connection(monkeypatch: pytest.MonkeyPatch):
    """Patches the test_connection method of the PostgresDatasource class to return True."""

    def _mock_test_connection(self: PostgresDatasource) -> bool:
        LOGGER.warning(
            f"Mocked {self.__class__.__name__}.test_connection() called and returning True"
        )
        return True

    monkeypatch.setattr(PostgresDatasource, "test_connection", _mock_test_connection)


@pytest.mark.postgresql
def test_construct_postgres_datasource(create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        assert source.name == "my_datasource"
        assert source.execution_engine_type is SqlAlchemyExecutionEngine
        assert source.assets == []


def assert_table_asset(
    asset: TableAsset,
    name: str,
    table_name: str,
    source: PostgresDatasource,
    batch_parameters: tuple[str, ...],
    partitioner: Optional[ColumnPartitioner] = None,
):
    assert asset.name == name
    assert asset.table_name == table_name
    assert asset.datasource == source
    assert asset.get_batch_parameters_keys(partitioner=partitioner) == batch_parameters


def assert_batch_request(
    batch_request, source_name: str, asset_name: str, options: BatchParameters
):
    assert batch_request.datasource_name == source_name
    assert batch_request.data_asset_name == asset_name
    assert batch_request.options == options


@pytest.mark.postgresql
def test_add_table_asset_with_partitioner(mocker, create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        mocker.patch("sqlalchemy.create_engine")
        mocker.patch("sqlalchemy.engine")
        inspect = mocker.patch("sqlalchemy.inspect")
        inspect.return_value = MockSaInspector()
        get_column_names = mocker.patch("tests.sqlalchemy_test_doubles.MockSaInspector.get_columns")
        get_column_names.return_value = [{"name": "my_col"}]
        has_table = mocker.patch("tests.sqlalchemy_test_doubles.MockSaInspector.has_table")
        has_table.return_value = True

        asset = source.add_table_asset(name="my_asset", table_name="my_table")
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        assert len(source.assets) == 1
        assert asset == source.assets[0]
        assert_table_asset(
            asset=asset,
            name="my_asset",
            table_name="my_table",
            source=source,
            batch_parameters=("year", "month"),
            partitioner=partitioner,
        )
        assert_batch_request(
            batch_request=asset.build_batch_request(
                options={"year": 2021, "month": 10}, partitioner=partitioner
            ),
            source_name="my_datasource",
            asset_name="my_asset",
            options={"year": 2021, "month": 10},
        )


@pytest.mark.postgresql
def test_add_table_asset_with_no_partitioner(mocker, create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        mocker.patch("sqlalchemy.create_engine")
        mocker.patch("sqlalchemy.engine")
        mocker.patch("sqlalchemy.inspect")

        asset = source.add_table_asset(name="my_asset", table_name="my_table")
        assert len(source.assets) == 1
        assert asset == source.assets[0]
        assert_table_asset(
            asset=asset,
            name="my_asset",
            table_name="my_table",
            source=source,
            batch_parameters=tuple(),
        )
        assert_batch_request(
            batch_request=asset.build_batch_request(),
            source_name="my_datasource",
            asset_name="my_asset",
            options={},
        )
        assert_batch_request(
            batch_request=asset.build_batch_request({}),
            source_name="my_datasource",
            asset_name="my_asset",
            options={},
        )


@pytest.mark.postgresql
def test_construct_table_asset_directly_with_no_partitioner(create_source):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        asset = TableAsset(name="my_asset", table_name="my_table")
        asset._datasource = source
        assert_batch_request(asset.build_batch_request(), "my_datasource", "my_asset", {})


def create_and_add_table_asset_without_testing_connection(
    source: PostgresDatasource,
    name: str,
    table_name: str,
) -> Tuple[PostgresDatasource, TableAsset]:
    table_asset = TableAsset(
        name=name,
        table_name=table_name,
    )
    # TODO: asset custom init
    table_asset._datasource = source
    source.assets.append(table_asset)
    return source, table_asset


def year_month_partitioner(column_name: str) -> SqlPartitionerYearAndMonth:
    return SqlPartitionerYearAndMonth(
        column_name=column_name,
        method_name="partition_on_year_and_month",
    )


@pytest.mark.postgresql
def test_construct_table_asset_directly_with_partitioner(create_source):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source,
            name="my_asset",
            table_name="my_table",
        )
        partitioner = ColumnPartitionerMonthly(column_name="col")
        assert_table_asset(
            asset,
            "my_asset",
            "my_table",
            source,
            ("year", "month"),
            partitioner=partitioner,
        )
        batch_parameters = {"year": 2022, "month": 10}
        assert_batch_request(
            asset.build_batch_request(options=batch_parameters, partitioner=partitioner),
            "my_datasource",
            "my_asset",
            batch_parameters,
        )


@pytest.mark.postgresql
def test_datasource_gets_batch_list_no_partitioner(empty_data_context, create_source):
    def validate_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        assert spec == {
            "batch_identifiers": {},
            "data_asset_name": "my_asset",
            "table_name": "my_table",
            "schema_name": None,
            "type": "table",
        }

    with create_source(
        validate_batch_spec=validate_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
    ) as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        source.get_batch_identifiers_list(asset.build_batch_request())


def assert_batch_specs_correct_with_year_month_partitioner_defaults(batch_specs):
    # We should have only collected 1 batch_spec
    assert len(batch_specs) == 1
    for year in _DEFAULT_TEST_YEARS:
        for month in _DEFAULT_TEST_MONTHS:
            spec = {
                "type": "table",
                "data_asset_name": "my_asset",
                "table_name": "my_table",
                "schema_name": None,
                "batch_identifiers": {"my_col": {"year": ANY, "month": ANY}},
                "partitioner_method": "partition_on_year_and_month",
                "partitioner_kwargs": {"column_name": "my_col"},
            }
            assert spec in batch_specs


def assert_batch_identifiers_correct_with_year_month_partitioner_defaults(
    batch_identifiers_list: list[dict],
):
    # We should have 1 batch_spec per (year, month) pair
    expected_batch_spec_num = len(_DEFAULT_TEST_YEARS) * len(_DEFAULT_TEST_MONTHS)
    assert len(batch_identifiers_list) == expected_batch_spec_num
    for year in _DEFAULT_TEST_YEARS:
        for month in _DEFAULT_TEST_MONTHS:
            assert {"year": year, "month": month} in batch_identifiers_list


@pytest.mark.postgresql
def test_datasource_gets_batch_partitioner_with_unspecified_batch_parameters(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
    ) as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        empty_batch_request = asset.build_batch_request(partitioner=partitioner)
        assert empty_batch_request.options == {}
        batch_identifiers_list = source.get_batch_identifiers_list(empty_batch_request)
        batch = source.get_batch(empty_batch_request)

        assert batch.metadata == {"month": 12, "year": 2022}
        assert_batch_specs_correct_with_year_month_partitioner_defaults(batch_specs)
        assert_batch_identifiers_correct_with_year_month_partitioner_defaults(
            [IDDict(bi) for bi in batch_identifiers_list]
        )


@pytest.mark.postgresql
def test_datasource_gets_batch_list_partitioner_with_batch_parameters_set_to_none(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
    ) as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        assert asset.get_batch_parameters_keys(partitioner=partitioner) == (
            "year",
            "month",
        )
        batch_request_with_none = asset.build_batch_request(
            options={"year": None, "month": None}, partitioner=partitioner
        )
        assert batch_request_with_none.options == {"year": None, "month": None}
        batch_identifiers_list = source.get_batch_identifiers_list(batch_request_with_none)
        batch = source.get_batch(batch_request_with_none)

        # We should have 1 batch_identifier per (year, month) pair

        assert batch.metadata == {"month": 12, "year": 2022}
        assert_batch_specs_correct_with_year_month_partitioner_defaults(batch_specs)
        assert_batch_identifiers_correct_with_year_month_partitioner_defaults(
            [IDDict(bi) for bi in batch_identifiers_list]
        )


@pytest.mark.postgresql
def test_datasource_gets_batch_partitioner_with_partially_specified_batch_parameters(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    batch_specs = []
    year = 2022

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year, "month": month} for month in list(range(1, 13))],
    ) as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        batch_request = asset.build_batch_request(options={"year": year}, partitioner=partitioner)
        identifiers_list = source.get_batch_identifiers_list(batch_request)
        batch = source.get_batch(batch_request)
        assert len(batch_specs) == 1
        assert batch_specs[0] == {
            "type": "table",
            "data_asset_name": "my_asset",
            "table_name": "my_table",
            "schema_name": None,
            "batch_identifiers": {"my_col": {"year": 2022, "month": 12}},
            "partitioner_method": "partition_on_year_and_month",
            "partitioner_kwargs": {"column_name": "my_col"},
        }

        assert len(identifiers_list) == len(_DEFAULT_TEST_MONTHS)
        for month in _DEFAULT_TEST_MONTHS:
            expected_metadata = IDDict({"month": month, "year": year})
            assert expected_metadata in identifiers_list
        assert batch.metadata == {"month": 12, "year": 2022}


@pytest.mark.postgresql
def test_datasource_gets_batch_with_fully_specified_batch_parameters(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    year = 2022
    month = 1

    def validate_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        assert spec == {
            "batch_identifiers": {"my_col": {"month": 1, "year": 2022}},
            "data_asset_name": "my_asset",
            "partitioner_kwargs": {"column_name": "my_col"},
            "partitioner_method": "partition_on_year_and_month",
            "table_name": "my_table",
            "schema_name": None,
            "type": "table",
        }

    with create_source(
        validate_batch_spec=validate_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"month": month, "year": year}],
    ) as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        batches = source.get_batch(
            asset.build_batch_request(
                options={"month": month, "year": year}, partitioner=partitioner
            )
        )
        assert batches.metadata == {"month": month, "year": year}


@pytest.mark.postgresql
def test_datasource_gets_nonexistent_asset(create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        with pytest.raises(LookupError):
            source.get_asset("my_asset")


@pytest.mark.postgresql
@pytest.mark.parametrize(
    [
        "PartitionerClass",
        "add_partitioner_kwargs",
        "batch_request_args",
    ],
    [
        # These bad parameters do not occur in the batch request params
        (ColumnPartitionerYearly, {"column_name": "my_col"}, ("bad", None, None)),
        (ColumnPartitionerYearly, {"column_name": "my_col"}, (None, "bad", None)),
        (ColumnPartitionerYearly, {"column_name": "my_col"}, ("bad", "bad", None)),
        # These bad parameters are request option parameters which depend on the partitioner.
        (
            ColumnPartitionerYearly,
            {"column_name": "my_col"},
            (None, None, {"bad": None}),
        ),
        (
            ColumnPartitionerMonthly,
            {"column_name": "my_col"},
            (None, None, {"bad": None}),
        ),
        (
            ColumnPartitionerDaily,
            {"column_name": "my_col"},
            (None, None, {"bad": None}),
        ),
        (
            PartitionerDatetimePart,
            {"column_name": "my_col", "datetime_parts": ["year"]},
            (None, None, {"bad": None}),
        ),
        (
            PartitionerColumnValue,
            {"column_name": "my_col"},
            (None, None, {"bad": None}),
        ),
        (
            PartitionerDividedInteger,
            {"column_name": "my_col", "divisor": 3},
            (None, None, {"bad": None}),
        ),
    ],
)
def test_bad_batch_request_passed_into_get_batch(
    create_source: CreateSourceFixture,
    PartitionerClass,
    add_partitioner_kwargs,
    batch_request_args,
):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        # We use a QueryAsset so no database connection checks are done.
        # We are mocking the database so these would fail without monkeypatching the
        # TableAsset
        asset = source.add_query_asset(name="query_asset", query="SELECT * FROM my_table")
        partitioner = PartitionerClass(**add_partitioner_kwargs)

        src, ast, op = batch_request_args
        batch_request = BatchRequest[ColumnPartitioner](
            datasource_name=src or source.name,
            data_asset_name=ast or asset.name,
            options=op or {},
            partitioner=partitioner,
        )
        with pytest.raises(
            (
                ge_exceptions.InvalidBatchRequestError,
                LookupError,
            )
        ):
            source.get_batch(batch_request)


@pytest.mark.postgresql
@pytest.mark.parametrize(
    "batch_parameters",
    [{}, {"year": 2021}, {"year": 2021, "month": 10}, {"year": None, "month": 10}],
)
def test_get_batch_with_good_batch_request(
    empty_data_context,
    create_source: CreateSourceFixture,
    batch_parameters,
):
    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
    ) as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        batch_request = BatchRequest(
            datasource_name=source.name,
            data_asset_name=asset.name,
            options=batch_parameters,
            partitioner=partitioner,
        )
        # No exception should get thrown
        asset.get_batch(batch_request)


@pytest.mark.postgresql
@pytest.mark.parametrize(
    "batch_request_args",
    [
        ("bad", None, None),
        (None, "bad", None),
        (None, None, {"bad": None}),
        ("bad", "bad", None),
    ],
)
def test_get_batch_with_malformed_batch_request(
    create_source: CreateSourceFixture, batch_request_args
):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        src, ast, op = batch_request_args
        batch_request = BatchRequest(
            datasource_name=src or source.name,
            data_asset_name=ast or asset.name,
            options=op or {},
            partitioner=partitioner,
        )
        with pytest.raises(ge_exceptions.InvalidBatchRequestError):
            asset.get_batch(batch_request)


@pytest.mark.postgresql
def test_get_bad_batch_request(create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        with pytest.raises(ge_exceptions.InvalidBatchRequestError):
            asset.build_batch_request(options={"invalid_key": None}, partitioner=partitioner)


@pytest.mark.postgresql
def test_data_source_json_has_properties(create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        # type should be in dumped json even if not explicitly set
        assert f'"type": "{asset.type}"'  # noqa: PLW0129 # FIXME CoP


@pytest.mark.postgresql
def test_data_source_yaml_has_properties(create_source: CreateSourceFixture):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            asset,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        source_str = source.__str__()
        # type should be in dumped str even if not explicitly set
        assert f"type: {asset.type}" in source_str


@pytest.mark.postgresql
def test_datasource_dict_has_properties(create_source):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        (
            source,  # noqa: PLW2901 # FIXME CoP
            _,
        ) = create_and_add_table_asset_without_testing_connection(
            source=source, name="my_asset", table_name="my_table"
        )
        source_dict = source.dict()
        # type should be in dumped dict even if not explicitly set
        assert (
            "type"
            in list(
                filter(
                    lambda element: element["name"] == "my_asset",
                    source_dict["assets"],
                )
            )[0]
        )


@pytest.mark.postgresql
@pytest.mark.parametrize(
    "connection_string",
    [
        "postgresql://userName:@hostname/dbName",
        "postgresql://userName@hostname",
        "postgresql://userName:password@hostname",
        "postgres://userName:@hostname",
        "postgresql+psycopg2://userName:@hostname",
        "postgresql+pg8000://userName:@hostname",
    ],
)
def test_validate_valid_postgres_connection_string(
    create_source: CreateSourceFixture, connection_string
):
    connection_string = "postgresql://userName:@hostname/dbName"
    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        connection_string=connection_string,
    ):
        # As long as no exception is thrown we consider this a pass. Pydantic normalizes the underlying  # noqa: E501 # FIXME CoP
        # connection string so a direct str comparison isn't possible.
        pass


@pytest.mark.postgresql
@pytest.mark.parametrize(
    "connection_string",
    [
        "postgresql://",
        "postgresql://username",
        "postgresql://username/dbName",
        "postgresql+invalid://",
    ],
)
def test_validate_invalid_postgres_connection_string(
    create_source: CreateSourceFixture, connection_string
):
    with pytest.raises(ValidationError):
        with create_source(
            validate_batch_spec=lambda _: None,
            dialect="postgresql",
            connection_string=connection_string,
        ):
            pass


def bad_connection_string_config() -> tuple[str, str, str]:
    connection_string = "postgresql+psycopg2://postgres:@localhost/bad_database"
    table_name = "good_table"
    schema_name = "good_schema"
    return (
        connection_string,
        table_name,
        schema_name,
    )


def bad_table_name_config() -> tuple[str, str, str]:
    connection_string = "postgresql+psycopg2://postgres:@localhost/test_ci"
    table_name = "bad_table"
    schema_name = "good_schema"
    return (
        connection_string,
        table_name,
        schema_name,
    )


def bad_schema_name_config() -> tuple[str, str, str]:
    connection_string = "postgresql+psycopg2://postgres:@localhost/test_ci"
    table_name = "good_table"
    schema_name = "bad_schema"
    return (
        connection_string,
        table_name,
        schema_name,
    )


@pytest.fixture(
    params=[
        bad_connection_string_config,
        bad_table_name_config,
        bad_schema_name_config,
    ]
)
def bad_configuration_datasource(
    request,
) -> PostgresDatasource:
    (
        connection_string,
        table_name,
        schema_name,
    ) = request.param()
    table_asset = TableAsset(
        name="table_asset",
        table_name=table_name,
        schema_name=schema_name,
    )
    return PostgresDatasource(
        name="postgres_datasource",
        connection_string=connection_string,
        assets=[
            table_asset,
        ],
    )


@pytest.fixture
def mock_create_engine(mocker: MockFixture) -> Mock:
    create_engine = mocker.patch("sqlalchemy.create_engine")
    create_engine.return_value = MockSaEngine(dialect=Dialect("postgresql"))
    return create_engine


@pytest.fixture
def mock_inspector(
    mocker: MockFixture,
) -> Generator[dict[Literal["schema_names", "table_names"], list[str]], None, None]:
    """
    Mock the inspector to return the schema and table names we want.
    Append the desired schema and table names (in tests) to the mock_inspector_returns dict.
    """
    mock_inspector_returns: dict[Literal["schema_names", "table_names"], list[str]] = {
        "schema_names": [],
        "table_names": [],
    }
    mock_inspector = MockSaInspector()

    def get_table_names(schema: str | None = None) -> list[str]:
        LOGGER.info("MockSaInspector.get_table_names() called")
        return mock_inspector_returns["table_names"]

    def get_schema_names() -> list[str]:
        LOGGER.info("MockSaInspector.get_schema_names() called")
        return mock_inspector_returns["schema_names"]

    def has_table(table_name: str, schema: str | None = None) -> bool:
        LOGGER.info("MockSaInspector.has_table() called")
        return table_name in mock_inspector_returns["table_names"]

    # directly patching the instance rather then using mocker.patch
    mock_inspector.get_schema_names = get_schema_names  # type: ignore[method-assign] # FIXME CoP
    mock_inspector.get_table_names = get_table_names  # type: ignore[method-assign] # FIXME CoP
    mock_inspector.has_table = has_table  # type: ignore[method-assign] # FIXME CoP

    inspect = mocker.patch("sqlalchemy.inspect")
    inspect.return_value = mock_inspector

    yield mock_inspector_returns


@pytest.fixture
def mock_connection_execute_failure(mocker: MockFixture) -> Mock:
    """
    Engine execute should raise an exception when called.
    """
    execute = mocker.patch("tests.sqlalchemy_test_doubles._MockConnection.execute")

    def execute_side_effect(*args, **kwargs):
        LOGGER.info("MockSaEngine.execute() called")
        raise SQLAlchemyError("Mocked exception")

    execute.side_effect = execute_side_effect
    return execute


@pytest.mark.postgresql
def test_test_connection_failures(
    mock_create_engine: Mock,
    mock_connection_execute_failure: Mock,
    mock_inspector: dict[Literal["schema_names", "table_names"], list[str]],
    bad_configuration_datasource: PostgresDatasource,
):
    mock_inspector["schema_names"].extend(["good_schema"])
    mock_inspector["table_names"].extend(["good_table", "bad_table"])

    with pytest.raises(TestConnectionError):
        bad_configuration_datasource.test_connection()


@pytest.mark.filesystem
def test_query_data_asset(empty_data_context, create_source):
    query = "SELECT * FROM my_table"

    def validate_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        assert spec == {
            "data_asset_name": "query_asset",
            "query": "SELECT * FROM my_table",
            "temp_table_schema_name": None,
            "batch_identifiers": {},
        }

    with create_source(
        validate_batch_spec=validate_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
    ) as source:
        asset = source.add_query_asset(name="query_asset", query="SELECT * FROM my_table")
        assert asset.name == "query_asset"
        assert asset.query.lower() == query.lower()
        source.get_batch(asset.build_batch_request())


@pytest.mark.postgresql
def test_non_select_query_data_asset(create_source):
    with create_source(validate_batch_spec=lambda _: None, dialect="postgresql") as source:
        with pytest.raises(ValueError):
            source.add_query_asset(name="query_asset", query="* FROM my_table")


@pytest.mark.filesystem
def test_adding_partitioner_persists_results(
    empty_data_context: FileDataContext,
    mock_test_connection,
):
    gx_yaml = pathlib.Path(empty_data_context.root_directory, FileDataContext.GX_YML).resolve(
        strict=True
    )

    empty_data_context.data_sources.add_postgres(
        name="my_datasource",
        connection_string="postgresql://postgres:@localhost/not_a_real_db",
    ).add_query_asset(name="my_asset", query="select * from table").add_batch_definition(
        name="my_batch_definition",
        partitioner=ColumnPartitionerYearly(column_name="my_col"),
        # We skip validating we can generate batches because table is fake and won't validate.
        validate_partitioner=False,
    )

    final_yaml: dict = YAMLHandler().load(  # type: ignore[assignment] # FIXME CoP
        gx_yaml.read_text(),
    )["fluent_datasources"]

    assert final_yaml["my_datasource"]["assets"]["my_asset"]["batch_definitions"][
        "my_batch_definition"
    ]["partitioner"]


@pytest.mark.postgresql
def test_partitioner_year(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    years = [2020, 2021]
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year} for year in years],
    ) as source:
        # We use a query asset because then we don't have to mock out db connection tests
        # in this unit test.
        asset = source.add_query_asset(name="my_asset", query="select * from table")
        partitioner = ColumnPartitionerYearly(column_name="my_col")
        batch_request = asset.build_batch_request(partitioner=partitioner)
        batches = source.get_batch_identifiers_list(batch_request)
        batch = source.get_batch(batch_request)
        assert len(batches) == len(years)
        for i, year in enumerate(years):
            assert "year" in batches[i]
            assert batches[i]["year"] == year

        assert len(batch_specs) == 1
        for spec in batch_specs:
            assert "partitioner_method" in spec
            assert spec["partitioner_method"] == "partition_on_year"
        assert batch.metadata == {"year": 2021}


@pytest.mark.postgresql
def test_partitioner_year_and_month(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    years = [2020, 2021]
    months = [6, 8, 9]
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[
            {"year": year, "month": month} for year in years for month in months
        ],
    ) as source:
        # We use a query asset because then we don't have to mock out db connection tests
        # in this unit test.
        asset = source.add_query_asset(name="my_asset", query="select * from table")
        partitioner = ColumnPartitionerMonthly(column_name="my_col")
        batch_request = asset.build_batch_request(partitioner=partitioner)
        batch_identifiers = source.get_batch_identifiers_list(batch_request)
        batch = source.get_batch(batch_request)
        assert len(batch_identifiers) == len(years) * len(months)
        for i, year in enumerate(years):
            for j, month in enumerate(months):
                batch_index = i * len(months) + j
                assert "year" in batch_identifiers[batch_index]
                assert "month" in batch_identifiers[batch_index]
                assert batch_identifiers[batch_index]["year"] == year
                assert batch_identifiers[batch_index]["month"] == month

        assert len(batch_identifiers) == len(years) * len(months)
        assert len(batch_specs) == 1
        assert "partitioner_method" in batch_specs[0]
        assert batch_specs[0]["partitioner_method"] == "partition_on_year_and_month"
        assert batch.metadata == {"year": 2021, "month": 9}


@pytest.mark.postgresql
def test_partitioner_year_and_month_and_day(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    years = [2020, 2021]
    months = [6, 8, 9]
    days = [1, 23, 24, 30]
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[
            {"year": year, "month": month, "day": day}
            for year in years
            for month in months
            for day in days
        ],
    ) as source:
        # We use a query asset because then we don't have to mock out db connection tests
        # in this unit test.
        asset = source.add_query_asset(
            name="my_asset",
            query="select * from table",
        )
        partitioner = ColumnPartitionerDaily(column_name="my_col")
        batch_request = asset.build_batch_request(partitioner=partitioner)
        batches = source.get_batch_identifiers_list(batch_request)
        batch = source.get_batch(batch_request)
        assert len(batches) == len(years) * len(months) * len(days)
        for i, year in enumerate(years):
            for j, month in enumerate(months):
                for k, day in enumerate(days):
                    batch_index = i * len(months) * len(days) + j * len(days) + k
                    assert "year" in batches[batch_index]
                    assert "month" in batches[batch_index]
                    assert "day" in batches[batch_index]
                    assert batches[batch_index]["year"] == year
                    assert batches[batch_index]["month"] == month
                    assert batches[batch_index]["day"] == day

        assert len(batch_specs) == 1
        assert "partitioner_method" in batch_specs[0]
        assert batch_specs[0]["partitioner_method"] == "partition_on_year_and_month_and_day"
        assert batch.metadata == {"year": 2021, "month": 9, "day": 30}


@pytest.mark.parametrize(
    ("sort_ascending", "expected_metadata"), [(True, {"year": 2021}), (False, {"year": 2020})]
)
@pytest.mark.postgresql
def test_get_batch_partitioner__sort_ascending_respected(
    empty_data_context,
    create_source: CreateSourceFixture,
    sort_ascending: bool,
    expected_metadata: dict,
):
    years = [2020, 2021]
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year} for year in years],
    ) as source:
        # We use a query asset because then we don't have to mock out db connection tests
        # in this unit test.
        asset = source.add_query_asset(
            name="my_asset",
            query="select * from table",
        )
        partitioner = ColumnPartitionerYearly(column_name="my_col", sort_ascending=sort_ascending)
        batch_request = asset.build_batch_request(partitioner=partitioner)
        batch = source.get_batch(batch_request)
        assert batch.metadata == expected_metadata


@pytest.mark.postgresql
def test_get_batch_raises_if_no_batches_available(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    batch_specs = []

    def collect_batch_spec(spec: SqlAlchemyDatasourceBatchSpec) -> None:
        batch_specs.append(spec)

    with create_source(
        validate_batch_spec=collect_batch_spec,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[],
    ) as source:
        # We use a query asset because then we don't have to mock out db connection tests
        # in this unit test.
        asset = source.add_query_asset(
            name="my_asset",
            query="select * from table",
        )
        partitioner = ColumnPartitionerYearly(column_name="my_col")
        batch_request = asset.build_batch_request(partitioner=partitioner, options={"year": 1995})
        with pytest.raises(NoAvailableBatchesError):
            source.get_batch(batch_request)


@pytest.mark.postgresql
@pytest.mark.parametrize(
    [
        "PartitionerClass",
        "partitioner_kwargs",
        "partitioner_query_responses",
        "all_batches_cnt",
        "specified_batch_request",
        "specified_batch_cnt",
        "last_specified_batch_metadata",
    ],
    [
        pytest.param(
            ColumnPartitionerYearly,
            {"column_name": "pickup_datetime"},
            [{"year": 2020}],
            1,
            {"year": 2020},
            1,
            {"year": 2020},
            id="year",
        ),
        pytest.param(
            ColumnPartitionerMonthly,
            {"column_name": "pickup_datetime"},
            [{"year": 2020, "month": 1}, {"year": 2020, "month": 2}],
            2,
            {"year": 2020, "month": 1},
            1,
            {"year": 2020, "month": 1},
            id="year_and_month",
        ),
        pytest.param(
            ColumnPartitionerDaily,
            {"column_name": "pickup_datetime"},
            [
                {"year": 2020, "month": 2, "day": 10},
                {"year": 2020, "month": 2, "day": 12},
            ],
            2,
            {"year": 2020, "month": 2, "day": 10},
            1,
            {"year": 2020, "month": 2, "day": 10},
            id="year_and_month_and_day",
        ),
        pytest.param(
            PartitionerDatetimePart,
            {
                "column_name": "pickup_datetime",
                "datetime_parts": ["year", "month", "day"],
            },
            [
                {"year": 2020, "month": 2, "day": 10},
                {"year": 2020, "month": 2, "day": 12},
            ],
            2,
            {"year": 2020, "month": 2},
            2,
            {"year": 2020, "month": 2, "day": 12},
            id="datetime_part",
        ),
        pytest.param(
            PartitionerColumnValue,
            {"column_name": "passenger_count"},
            [(1,), (None,), (2,)],
            3,
            {"passenger_count": 2},
            1,
            {"passenger_count": 2},
            id="column_value",
        ),
        pytest.param(
            PartitionerDividedInteger,
            {"column_name": "passenger_count", "divisor": 3},
            [(1,), (2,)],
            2,
            {"quotient": 2},
            1,
            {"quotient": 2},
            id="divisor",
        ),
        pytest.param(
            PartitionerModInteger,
            {"column_name": "passenger_count", "mod": 3},
            [(1,), (2,)],
            2,
            {"remainder": 2},
            1,
            {"remainder": 2},
            id="mod_integer",
        ),
        pytest.param(
            PartitionerMultiColumnValue,
            {"column_names": ["passenger_count", "payment_type"]},
            # These types are (passenger_count, payment_type), that is in column_names order.
            # datetime partitioners return dicts while all other partitioners return tuples.
            [(3, 1), (1, 1), (1, 2)],
            3,
            {"passenger_count": 1},
            2,
            {"passenger_count": 1, "payment_type": 2},
            id="multi_column_values",
        ),
    ],
)
def test_partitioner(
    empty_data_context,
    create_source: CreateSourceFixture,
    PartitionerClass,
    partitioner_kwargs,
    partitioner_query_responses,
    all_batches_cnt,
    specified_batch_request,
    specified_batch_cnt,
    last_specified_batch_metadata,
):
    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[response for response in partitioner_query_responses],
    ) as source:
        asset = source.add_query_asset(name="query_asset", query="SELECT * from table")
        partitioner = PartitionerClass(**partitioner_kwargs)
        # Test getting all batch itentifiers
        all_batches = asset.get_batch_identifiers_list(
            asset.build_batch_request(partitioner=partitioner)
        )
        assert len(all_batches) == all_batches_cnt
        # Test getting specified batches
        batch_request = asset.build_batch_request(specified_batch_request, partitioner=partitioner)
        specified_batches = asset.get_batch_identifiers_list(batch_request)
        assert len(specified_batches) == specified_batch_cnt
        assert asset.get_batch(batch_request).metadata == last_specified_batch_metadata


@pytest.mark.postgresql
def test_sorting_none_in_metadata(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    years = [None, 2020, 2021]

    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year} for year in years],
    ) as source:
        # We use a query asset because then we don't have to mock out db connection tests
        # in this unit test.
        asset = source.add_query_asset(name="my_asset", query="select * from table")
        partitioner = ColumnPartitionerYearly(column_name="my_col", sort_ascending=False)
        batch_request = asset.build_batch_request(partitioner=partitioner)
        batches = source.get_batch_identifiers_list(batch_request)
        assert len(batches) == len(years)
        assert asset.get_batch(batch_request).metadata["year"] is None


@pytest.mark.postgresql
def test_create_temp_table(empty_data_context, create_source):
    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
        create_temp_table=False,
    ) as source:
        assert source.create_temp_table is False
        asset = source.add_query_asset(name="query_asset", query="SELECT * from table")
        _ = asset.get_batch(asset.build_batch_request())
        assert source._execution_engine._create_temp_table is False


@pytest.mark.postgresql
def test_add_postgres_query_asset_with_batch_metadata(
    empty_data_context,
    create_source: CreateSourceFixture,
):
    my_config_variables = {"pipeline_filename": __file__}
    empty_data_context.config_variables.update(my_config_variables)

    years = [2021, 2022]
    asset_specified_metadata = {
        "pipeline_name": "my_pipeline",
        "no_curly_pipeline_filename": "$pipeline_filename",
        "curly_pipeline_filename": "${pipeline_filename}",
    }

    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year} for year in years],
    ) as source:
        asset = source.add_query_asset(
            name="query_asset",
            query="SELECT * FROM my_table",
            batch_metadata=asset_specified_metadata,
        )
        assert asset.batch_metadata == asset_specified_metadata
        partitioner = ColumnPartitionerYearly(column_name="col")
        batches = source.get_batch_identifiers_list(
            asset.build_batch_request(partitioner=partitioner)
        )
        assert len(batches) == len(years)
        substituted_batch_metadata: BatchMetadata = copy.deepcopy(asset_specified_metadata)
        substituted_batch_metadata.update(
            {
                "no_curly_pipeline_filename": __file__,
                "curly_pipeline_filename": __file__,
            }
        )
        for i, year in enumerate(years):
            substituted_batch_metadata["year"] = year
            assert batches[i] == substituted_batch_metadata


@pytest.mark.postgresql
def test_add_postgres_table_asset_with_batch_metadata(
    empty_data_context, create_source: CreateSourceFixture, monkeypatch
):
    my_config_variables = {"pipeline_filename": __file__}
    empty_data_context.config_variables.update(my_config_variables)

    monkeypatch.setattr(TableAsset, "test_connection", lambda _: None)
    years = [2021, 2022]
    asset_specified_metadata = {
        "pipeline_name": "my_pipeline",
        "no_curly_pipeline_filename": "$pipeline_filename",
        "curly_pipeline_filename": "${pipeline_filename}",
    }

    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year} for year in years],
    ) as source:
        asset = source.add_table_asset(
            name="query_asset",
            table_name="my_table",
            batch_metadata=asset_specified_metadata,
        )
        assert asset.batch_metadata == asset_specified_metadata
        partitioner = ColumnPartitionerYearly(column_name="my_col")
        batches = source.get_batch_identifiers_list(
            asset.build_batch_request(partitioner=partitioner)
        )
        assert len(batches) == len(years)
        substituted_batch_metadata: BatchMetadata = copy.deepcopy(asset_specified_metadata)
        substituted_batch_metadata.update(
            {
                "no_curly_pipeline_filename": __file__,
                "curly_pipeline_filename": __file__,
            }
        )
        for i, year in enumerate(years):
            substituted_batch_metadata["year"] = year
            assert batches[i] == substituted_batch_metadata



from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict, Optional, TypeVar, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.id_dict import IDDict
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.types import SerializableDictDot, SerializableDotDict
from great_expectations.util import (
    convert_to_json_serializable,  # noqa: TID251 # FIXME CoP
    deep_filter_properties_iterable,
    is_candidate_subset_of_target,
)

INFERRED_SEMANTIC_TYPE_KEY: str = "inferred_semantic_domain_type"

T = TypeVar("T")


class SemanticDomainTypes(Enum):
    NUMERIC = "numeric"
    TEXT = "text"
    LOGIC = "logic"
    DATETIME = "datetime"
    BINARY = "binary"
    CURRENCY = "currency"
    IDENTIFIER = "identifier"
    MISCELLANEOUS = "miscellaneous"
    UNKNOWN = "unknown"


@dataclass
class InferredSemanticDomainType(SerializableDictDot):
    semantic_domain_type: Optional[Union[str, SemanticDomainTypes]] = None
    details: Optional[Dict[str, Any]] = None

    @override
    def to_dict(self) -> dict:
        return asdict(self)

    @override
    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(data=self.to_dict())


class DomainKwargs(SerializableDotDict):
    def to_dict(self) -> dict:
        return dict(self)

    @override
    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(data=self.to_dict())


class Domain(SerializableDotDict):
    # Adding an explicit constructor to highlight the specific properties that will be used.
    def __init__(  # noqa: C901 #  too complex
        self,
        domain_type: Union[str, MetricDomainTypes],
        domain_kwargs: Optional[Union[Dict[str, Any], DomainKwargs]] = None,
        details: Optional[Dict[str, Any]] = None,
        rule_name: Optional[str] = None,
    ) -> None:
        if isinstance(domain_type, str):
            try:
                domain_type = MetricDomainTypes(domain_type.lower())
            except (TypeError, KeyError) as e:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    f""" {e}: Cannot instantiate Domain (domain_type "{domain_type!s}" of type \
"{type(domain_type)!s}" is not supported).
"""
                )
        elif not isinstance(domain_type, MetricDomainTypes):
            raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                f"""Cannot instantiate Domain (domain_type "{domain_type!s}" of type "{type(domain_type)!s}" is \
not supported).
"""  # noqa: E501 # FIXME CoP
            )

        if domain_kwargs is None:
            domain_kwargs = DomainKwargs({})
        elif isinstance(domain_kwargs, dict):
            domain_kwargs = DomainKwargs(domain_kwargs)

        domain_kwargs_dot_dict: DomainKwargs = deep_convert_properties_iterable_to_domain_kwargs(
            source=domain_kwargs
        )

        if details is None:
            details = {}

        inferred_semantic_domain_type: Optional[Dict[str, Union[str, SemanticDomainTypes]]] = (
            details.get(INFERRED_SEMANTIC_TYPE_KEY)
        )
        if inferred_semantic_domain_type:
            semantic_domain_key: str
            metric_domain_key: str
            metric_domain_value: Any
            is_consistent: bool
            for semantic_domain_key in inferred_semantic_domain_type:
                is_consistent = False
                for (
                    metric_domain_key,
                    metric_domain_value,
                ) in domain_kwargs_dot_dict.items():
                    if (
                        isinstance(metric_domain_value, (list, set, tuple))
                        and semantic_domain_key in metric_domain_value
                    ) or (semantic_domain_key == metric_domain_value):
                        is_consistent = True
                        break

                if not is_consistent:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        f"""Cannot instantiate Domain (domain_type "{domain_type!s}" of type \
"{type(domain_type)!s}" -- key "{semantic_domain_key}", detected in "{INFERRED_SEMANTIC_TYPE_KEY}" dictionary, does \
not exist as value of appropriate key in "domain_kwargs" dictionary.
"""  # noqa: E501 # FIXME CoP
                    )

        super().__init__(
            domain_type=domain_type,
            domain_kwargs=domain_kwargs_dot_dict,
            details=details,
            rule_name=rule_name,
        )

    @override
    def __repr__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    @override
    def __str__(self):
        return self.__repr__()

    @override
    def __eq__(self, other):
        return (other is not None) and (
            (hasattr(other, "to_json_dict") and self.to_json_dict() == other.to_json_dict())
            or (
                isinstance(other, dict)
                and deep_filter_properties_iterable(
                    properties=self.to_json_dict(), clean_falsy=True
                )
                == deep_filter_properties_iterable(properties=other, clean_falsy=True)
            )
            or (self.__str__() == str(other))
        )

    @override
    def __ne__(self, other):
        return not self.__eq__(other=other)

    @override
    def __hash__(self) -> int:  # type: ignore[override] # FIXME CoP
        """Overrides the default implementation"""
        _result_hash: int = hash(self.id)
        return _result_hash

    def is_superset(self, other: Domain) -> bool:
        """Determines if other "Domain" object (provided as argument) is contained within this "Domain" object."""  # noqa: E501 # FIXME CoP
        if other is None:
            return True

        return other.is_subset(other=self)

    def is_subset(self, other: Domain) -> bool:
        """Determines if this "Domain" object is contained within other "Domain" object (provided as argument)."""  # noqa: E501 # FIXME CoP
        if other is None:
            return False

        this_json_dict: dict = self.to_json_dict()
        other_json_dict: dict = other.to_json_dict()

        return is_candidate_subset_of_target(candidate=this_json_dict, target=other_json_dict)

    # Adding this property for convenience (also, in the future, arguments may not be all set to their default values).  # noqa: E501 # FIXME CoP
    @property
    def id(self) -> str:
        return str(IDDict(self.to_json_dict()).to_id())

    @override
    def to_json_dict(self) -> dict:
        details: dict = {}

        key: str
        value: Any
        for key, value in self["details"].items():
            if value:
                if key == INFERRED_SEMANTIC_TYPE_KEY:
                    column_name: str
                    semantic_type: Union[str, SemanticDomainTypes]
                    value = {  # noqa: PLW2901 # FIXME CoP
                        column_name: SemanticDomainTypes(semantic_type.lower()).value
                        if isinstance(semantic_type, str)
                        else semantic_type.value
                        for column_name, semantic_type in value.items()
                    }

            details[key] = convert_to_json_serializable(data=value)

        json_dict: dict = {
            "domain_type": self["domain_type"].value,
            "domain_kwargs": self["domain_kwargs"].to_json_dict(),
            "details": details,
            "rule_name": self["rule_name"],
        }
        json_dict = convert_to_json_serializable(data=json_dict)

        return deep_filter_properties_iterable(properties=json_dict, clean_falsy=True)


def deep_convert_properties_iterable_to_domain_kwargs(
    source: Union[T, dict],
) -> Union[T, DomainKwargs]:
    if isinstance(source, dict):
        return _deep_convert_properties_iterable_to_domain_kwargs(source=DomainKwargs(source))

    # Must allow for non-dictionary source types, since their internal nested structures may contain dictionaries.  # noqa: E501 # FIXME CoP
    if isinstance(source, (list, set, tuple)):
        data_type: type = type(source)

        element: Any
        return data_type(
            [
                deep_convert_properties_iterable_to_domain_kwargs(source=element)
                for element in source
            ]
        )

    return source


def _deep_convert_properties_iterable_to_domain_kwargs(source: dict) -> DomainKwargs:
    key: str
    value: Any
    for key, value in source.items():
        if isinstance(value, dict):
            source[key] = _deep_convert_properties_iterable_to_domain_kwargs(source=value)
        elif isinstance(value, (list, set, tuple)):
            data_type: type = type(value)

            element: Any
            source[key] = data_type(
                [
                    deep_convert_properties_iterable_to_domain_kwargs(source=element)
                    for element in value
                ]
            )

    return DomainKwargs(source)



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any
from unittest import mock

import pytest

from great_expectations.checkpoint import SlackNotificationAction
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.data_context_key import StringKey
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context import AbstractDataContext
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.context_factory import set_context
from great_expectations.data_context.store.checkpoint_store import CheckpointStore
from great_expectations.data_context.types.resource_identifiers import GXCloudIdentifier

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tests.datasource.fluent._fake_cloud_api import CloudDetails


@pytest.fixture
def ephemeral_store():
    return CheckpointStore(store_name="ephemeral_checkpoint_store")


@pytest.fixture
def file_backed_store(tmp_path):
    base_directory = tmp_path / "base_dir"
    return CheckpointStore(
        store_name="file_backed_checkpoint_store",
        store_backend={
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": base_directory,
        },
    )


@pytest.fixture
def cloud_backed_store(cloud_details: CloudDetails):
    return CheckpointStore(
        store_name="cloud_backed_checkpoint_store",
        store_backend={
            "class_name": "GXCloudStoreBackend",
            "ge_cloud_resource_type": GXCloudRESTResource.CHECKPOINT,
            "ge_cloud_credentials": {
                "access_token": cloud_details.access_token,
                "organization_id": cloud_details.org_id,
            },
        },
    )


@pytest.fixture
def mock_checkpoint_json() -> dict:
    return {
        "name": "my_checkpoint",
        "validation_definitions": [
            {"name": "my_first_validation", "id": "a58816-64c8-46cb-8f7e-03c12cea1d67"},
            {"name": "my_second_validation", "id": "139ab16-64c8-46cb-8f7e-03c12cea1d67"},
        ],
        "actions": [
            {
                "name": "my_slack_action",
                "slack_webhook": "https://hooks.slack.com/services/ABC123/DEF456/XYZ789",
                "notify_on": "all",
                "notify_with": ["my_data_docs_site"],
                "renderer": {
                    "class_name": "SlackRenderer",
                },
            }
        ],
        "result_format": "SUMMARY",
        "id": None,
    }


@pytest.fixture
def mock_checkpoint_dict(mocker, mock_checkpoint_json: dict) -> dict:
    context = mocker.Mock(spec=AbstractDataContext)
    set_context(context)

    data = BatchDefinition[None](name="my_batch_config")
    suite = ExpectationSuite(name="my_suite")

    return {
        "name": mock_checkpoint_json["name"],
        "validation_definitions": [
            ValidationDefinition(
                **mock_checkpoint_json["validation_definitions"][0], data=data, suite=suite
            ),
            ValidationDefinition(
                **mock_checkpoint_json["validation_definitions"][1], data=data, suite=suite
            ),
        ],
        "actions": [
            SlackNotificationAction(**mock_checkpoint_json["actions"][0]),
        ],
        "result_format": mock_checkpoint_json["result_format"],
        "id": mock_checkpoint_json["id"],
    }


@pytest.fixture
def checkpoint(
    mocker: MockerFixture, mock_checkpoint_json: dict, mock_checkpoint_dict: dict
) -> CheckpointStore:
    cp = mocker.Mock(spec=Checkpoint, name="my_checkpoint", id=None)
    cp.json.return_value = json.dumps(mock_checkpoint_json)
    cp.dict.return_value = mock_checkpoint_dict
    return cp


@pytest.mark.parametrize("store_fixture", ["ephemeral_store", "file_backed_store"])
@pytest.mark.unit
def test_add(request, store_fixture: str, checkpoint: Checkpoint):
    store: CheckpointStore = request.getfixturevalue(store_fixture)
    key = store.get_key(name="my_checkpoint")

    assert not checkpoint.id
    store.add(key=key, value=checkpoint)

    assert checkpoint.id


@pytest.mark.cloud
def test_add_cloud(cloud_backed_store: CheckpointStore, checkpoint: Checkpoint):
    store = cloud_backed_store

    id = "5a8ada9f-5b71-461b-b1af-f1d93602a156"

    name = "my_checkpoint"
    key = GXCloudIdentifier(
        resource_type=GXCloudRESTResource.CHECKPOINT,
        id=id,
        resource_name=name,
    )

    with mock.patch("requests.Session.put", autospec=True) as mock_put:
        store.add(key=key, value=checkpoint)

    mock_put.assert_called_once_with(
        mock.ANY,  # requests Session
        f"https://api.greatexpectations.io/api/v1/organizations/12345678-1234-5678-1234-567812345678/checkpoints/{id}",
        json={
            "data": {
                "name": name,
                "id": None,
                "result_format": "SUMMARY",
                "validation_definitions": [
                    {
                        "id": "a58816-64c8-46cb-8f7e-03c12cea1d67",
                        "name": "my_first_validation",
                    },
                    {
                        "id": "139ab16-64c8-46cb-8f7e-03c12cea1d67",
                        "name": "my_second_validation",
                    },
                ],
                "actions": [
                    {
                        "name": "my_slack_action",
                        "notify_on": "all",
                        "notify_with": ["my_data_docs_site"],
                        "renderer": {"class_name": "SlackRenderer"},
                        "slack_webhook": "https://hooks.slack.com/services/ABC123/DEF456/XYZ789",
                    },
                ],
            }
        },
    )


@pytest.mark.parametrize("store_fixture", ["ephemeral_store", "file_backed_store"])
@pytest.mark.unit
def test_get_key(request, store_fixture: str):
    store: CheckpointStore = request.getfixturevalue(store_fixture)

    name = "my_checkpoint"
    assert store.get_key(name=name) == StringKey(key=name)


@pytest.mark.cloud
def test_get_key_cloud(cloud_backed_store: CheckpointStore):
    key = cloud_backed_store.get_key(name="my_checkpoint")
    assert key.resource_type == GXCloudRESTResource.CHECKPOINT  # type: ignore[union-attr] # FIXME CoP
    assert key.resource_name == "my_checkpoint"  # type: ignore[union-attr] # FIXME CoP


def _create_checkpoint_config(name: str, id: str) -> dict[str, Any]:
    return {
        "name": name,
        "validation_definitions": [
            {"name": "my_first_validation", "id": "a58816-64c8-46cb-8f7e-03c12cea1d67"},
            {"name": "my_second_validation", "id": "139ab16-64c8-46cb-8f7e-03c12cea1d67"},
        ],
        "actions": [
            {
                "name": "my_slack_action",
                "slack_webhook": "https://hooks.slack.com/services/ABC123/DEF456/XYZ789",
                "notify_on": "all",
                "notify_with": ["my_data_docs_site"],
                "renderer": {
                    "class_name": "SlackRenderer",
                },
            }
        ],
        "result_format": "SUMMARY",
        "id": id,
    }


_CHECKPOINT_ID = "a4sdfd-64c8-46cb-8f7e-03c12cea1d67"
_CHECKPOINT_CONFIG = _create_checkpoint_config("my_checkpoint", _CHECKPOINT_ID)


@pytest.mark.unit
@pytest.mark.parametrize(
    "response_json",
    [
        pytest.param(
            {
                "data": {
                    **_CHECKPOINT_CONFIG,
                    "id": _CHECKPOINT_ID,
                }
            },
            id="single_checkpoint_config",
        ),
        pytest.param(
            {
                "data": [
                    {
                        **_CHECKPOINT_CONFIG,
                        "id": _CHECKPOINT_ID,
                    }
                ]
            },
            id="list_with_single_checkpoint_config",
        ),
    ],
)
def test_gx_cloud_response_json_to_object_dict_success(response_json: dict):
    actual = CheckpointStore.gx_cloud_response_json_to_object_dict(response_json)
    expected = {**_CHECKPOINT_CONFIG, "id": _CHECKPOINT_ID}
    assert actual == expected


@pytest.mark.unit
def test_gx_cloud_response_json_to_object_collection():
    id_a = "a4sdfd-64c8-46cb-8f7e-03c12cea1d67"
    id_b = "a4sdfd-64c8-46cb-8f7e-03c12cea1d68"
    config_a = _create_checkpoint_config("my_checkpoint_a", "something else?")
    config_b = _create_checkpoint_config("my_checkpoint_b", id_b)
    response_json = {
        "data": [{**config_a, "id": id_a}, {**config_b, "id": id_b}],
    }

    result = CheckpointStore.gx_cloud_response_json_to_object_collection(response_json)

    expected = [{**config_a, "id": id_a}, {**config_b, "id": id_b}]
    assert result == expected


@pytest.mark.cloud
@pytest.mark.parametrize(
    "response_json, error_substring",
    [
        pytest.param(
            {
                "data": [],
            },
            "Cannot parse empty data",
            id="empty_list",
        ),
        pytest.param(
            {
                "data": [
                    {
                        "id": _CHECKPOINT_ID,
                        "attributes": {
                            "checkpoint_config": _CHECKPOINT_CONFIG,
                        },
                    },
                    {
                        "id": _CHECKPOINT_ID,
                        "attributes": {
                            "checkpoint_config": _CHECKPOINT_CONFIG,
                        },
                    },
                ],
            },
            "Cannot parse multiple items",
            id="list_with_multiple_checkpoint_configs",
        ),
    ],
)
def test_gx_cloud_response_json_to_object_dict_failure(response_json: dict, error_substring: str):
    with pytest.raises(ValueError, match=f"{error_substring}*."):
        CheckpointStore.gx_cloud_response_json_to_object_dict(response_json)


@pytest.mark.unit
def test_update_failure_wraps_store_backend_error(
    ephemeral_store: CheckpointStore, checkpoint: Checkpoint
):
    key = ephemeral_store.get_key(name="my_nonexistant_checkpoint")

    with pytest.raises(ValueError) as e:
        ephemeral_store.update(key=key, value=checkpoint)

    assert "Could not update Checkpoint" in str(e.value)



"""
This file is intended to
1. test the basic behavior of SerializableDictDot, in combination with @dataclass, and
2. provides examples of best practice for working with typed objects within the Great Expectations codebase
"""  # noqa: E501 # FIXME CoP

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Tuple

import pytest
from pytest import raises

from great_expectations.types import DictDot, SerializableDictDot


@dataclass
class MyClassA(SerializableDictDot):
    foo: str
    bar: int


@dataclass
class MyClassB(MyClassA):
    baz: List[str]
    qux: Optional[int] = None
    quux: int = 42

    @property
    def num_bazzes(self):
        return len(self.baz)


class MyEnum(Enum):
    VALUE_X = "x"
    VALUE_Y = "y"
    VALUE_Z = "z"


@dataclass
class MyClassC(SerializableDictDot):
    alpha_var: int
    beta_var: MyEnum
    A_list: List[MyClassA]
    B_list: List[MyClassB]
    enum_list: List[MyEnum] = field(default_factory=list)
    some_tuple: Optional[Tuple[MyClassA, MyClassB]] = None

    @property
    def num_As(self):
        return len(self.A_list)

    @property
    def num_Bs(self):
        return len(self.B_list)


@pytest.mark.unit
def test_access_using_dict_notation():
    "Keys can be accessed using dict notation"
    my_A = MyClassA(
        **{
            "foo": "a string",
            "bar": 1,
        }
    )
    assert my_A["foo"] == "a string"
    assert my_A["bar"] == 1


@pytest.mark.unit
def test_has_keys():
    "the .keys method works"
    my_A = MyClassA(
        **{
            "foo": "a string",
            "bar": 1,
        }
    )
    assert set(my_A.keys()) == {"foo", "bar"}


@pytest.mark.unit
def test_has_items():
    "the .items method works"
    my_A = MyClassA(
        **{
            "foo": "a string",
            "bar": 1,
        }
    )
    items = list(my_A.items())
    assert items == [("foo", "a string"), ("bar", 1)]


@pytest.mark.unit
def test_has_to_raw_dict():
    "the .to_raw_dict method works"
    my_A = MyClassA(
        **{
            "foo": "a string",
            "bar": 1,
        }
    )
    assert my_A.to_raw_dict() == {
        "foo": "a string",
        "bar": 1,
    }


@pytest.mark.skip(reason="Not sure what our preferred pattern for this is")
@pytest.mark.unit
def test_incorrect_type():
    "Throws an error if instantiated with an incorrect type"
    MyClassA(**{"foo": "a string", "bar": "SHOULD BE AN INT", "baz": ["a", "b", "c"]})


@pytest.mark.unit
def test_renders_to_a_useful_str():
    "Renders to a string that exposes the internals of the object"
    assert (
        MyClassA(
            foo="a string",
            bar=1,
        ).__str__()
        == """MyClassA(foo='a string', bar=1)"""
    )


@pytest.mark.skip(reason="hmmm. We should be able to make this work by default")
@pytest.mark.unit
def test_is_json_serializable():
    assert MyClassA(
        foo="a string",
        bar=1,
    ).to_json_dict() == {"foo": "a string", "bar": 1}


@pytest.mark.unit
def test_missing_keys():
    "Throws a TypeError if instantiated with missing arguments"

    with raises(TypeError):
        MyClassA()

    with raises(TypeError):
        MyClassA(
            foo="a string",
        )

    with raises(TypeError):
        MyClassA(bar=1)


@pytest.mark.unit
def test_extra_keys():
    "Throws a TypeError if instantiated with extra arguments"

    with raises(TypeError):
        MyClassA(**{"foo": "a string", "bar": 1, "baz": ["a", "b", "c"]})


@pytest.mark.unit
def test_update_after_instantiation():
    "Can be updated after instantiation, using both dot and dict notation"

    my_A = MyClassA(
        **{
            "foo": "a string",
            "bar": 1,
        }
    )
    assert my_A["foo"] == "a string"
    assert my_A["bar"] == 1

    my_A.foo = "different string"
    assert my_A["foo"] == "different string"
    assert my_A.foo == "different string"

    my_A["foo"] = "a third string"
    assert my_A["foo"] == "a third string"
    assert my_A.foo == "a third string"


@pytest.mark.unit
def test_can_be_subclassed():
    my_B = MyClassB(
        foo="a string",
        bar=1,
        baz=["a", "b", "c"],
        qux=-100,
        quux=43,
    )

    assert my_B["foo"] == "a string"
    assert my_B["bar"] == 1
    assert my_B["baz"] == ["a", "b", "c"]
    assert my_B["qux"] == -100
    assert my_B["quux"] == 43

    assert my_B.foo == "a string"
    assert my_B.bar == 1
    assert my_B.baz == ["a", "b", "c"]
    assert my_B.qux == -100
    assert my_B.quux == 43

    my_B = MyClassB(
        **{
            "foo": "a string",
            "bar": 1,
            "baz": ["a", "b", "c"],
            "qux": -100,
            "quux": 43,
        }
    )

    assert my_B["foo"] == "a string"
    assert my_B["bar"] == 1
    assert my_B["baz"] == ["a", "b", "c"]
    assert my_B["qux"] == -100
    assert my_B["quux"] == 43

    assert my_B.foo == "a string"
    assert my_B.bar == 1
    assert my_B.baz == ["a", "b", "c"]
    assert my_B.qux == -100
    assert my_B.quux == 43


@pytest.mark.unit
def test_can_have_derived_properties():
    "Can have derived properties"
    my_B = MyClassB(
        **{
            "foo": "a string",
            "bar": 1,
            "baz": ["a", "b", "c"],
            "qux": -100,
        }
    )
    assert my_B.num_bazzes == 3


@pytest.mark.unit
def test_can_have_optional_arguments():
    "Can have optional arguments"
    my_B = MyClassB(
        **{
            "foo": "a string",
            "bar": 1,
            "baz": ["a", "b", "c"],
            # "qux": -100, #Optional property
        }
    )
    assert my_B.qux is None


@pytest.mark.unit
def test_can_have_default_values():
    "Can have default values"
    my_B = MyClassB(
        **{
            "foo": "a string",
            "bar": 1,
            "baz": ["a", "b", "c"],
            "qux": -100,  # Optional property
        }
    )
    assert my_B.quux == 42


@pytest.mark.unit
def test_can_use_positional_arguments():
    "Can use a normal mix of positional and keyword arguments"
    my_B = MyClassB(
        "a string",
        1,
        baz=["a", "b", "c"],
        qux=-100,
    )
    assert my_B.foo == "a string"
    assert my_B.bar == 1
    assert my_B.baz == ["a", "b", "c"]
    assert my_B.qux == -100

    my_B = MyClassB(
        qux=-100,
        baz=["a", "b", "c"],
        bar=1,
        foo="a string",
    )
    assert my_B.foo == "a string"
    assert my_B.bar == 1
    assert my_B.baz == ["a", "b", "c"]
    assert my_B.qux == -100


@pytest.mark.unit
def test_can_be_nested():
    "Objects of this type can be nested"

    my_C = MyClassC(
        alpha_var=20,
        beta_var=MyEnum("x"),
        A_list=[
            MyClassA(
                foo="A-1",
                bar=101,
            ),
            MyClassA(
                **{
                    "foo": "A-2",
                    "bar": 102,
                }
            ),
        ],
        B_list=[
            MyClassB(
                **{
                    "foo": "B-1",
                    "bar": 201,
                    "baz": ["a", "b", "c"],
                    "qux": -100,
                    "quux": 43,
                }
            )
        ],
    )

    # Demonstrate that we can access sub-objects using a mix of dict and dot notation:
    assert my_C.A_list[0].foo == "A-1"
    assert my_C["A_list"][1].bar == 102
    assert my_C["B_list"][0]["quux"] == 43

    # Note: we don't currently support dot notation access within lists: `assert my_C["A_list"].1.bar == 102`  # noqa: E501 # FIXME CoP

    # Demonstrate that we can access Enum sub-objects
    assert my_C["beta_var"] == MyEnum("x")


@pytest.mark.unit
def test_to_raw_dict_works_recursively():
    "the .to_raw_dict method works recursively on both DotDicts and Enums"

    my_C = MyClassC(
        alpha_var=20,
        beta_var=MyEnum("x"),
        A_list=[
            MyClassA(
                foo="A-1",
                bar=101,
            ),
            MyClassA(
                **{
                    "foo": "A-2",
                    "bar": 102,
                }
            ),
        ],
        B_list=[
            MyClassB(
                **{
                    "foo": "B-1",
                    "bar": 201,
                    "baz": ["a", "b", "c"],
                    "qux": -100,
                    "quux": 43,
                }
            )
        ],
        enum_list=[
            MyEnum("x"),
            MyEnum("y"),
            MyEnum("z"),
        ],
        some_tuple=(
            MyClassA(
                foo="A-1",
                bar=101,
            ),
            MyClassB(
                foo="B-1",
                bar=201,
                baz=["a", "b", "c"],
                qux=-100,
                quux=43,
            ),
        ),
    )

    C_dict = my_C.to_raw_dict()

    # Make sure it's a dictionary, not a DictDot
    assert type(C_dict) == dict  # noqa: E721 # FIXME CoP
    assert isinstance(C_dict, DictDot) is False
    # Dictionaries don't support dot notation.
    with raises(AttributeError):
        C_dict.A_list  # noqa: B018 # FIXME CoP

    assert type(C_dict["A_list"][0]) == dict  # noqa: E721 # FIXME CoP
    assert type(C_dict["B_list"][0]) == dict  # noqa: E721 # FIXME CoP

    assert C_dict == {
        "alpha_var": 20,
        "beta_var": "x",
        "A_list": [
            {
                "foo": "A-1",
                "bar": 101,
            },
            {
                "foo": "A-2",
                "bar": 102,
            },
        ],
        "B_list": [
            {
                "foo": "B-1",
                "bar": 201,
                "baz": ["a", "b", "c"],
                "qux": -100,
                "quux": 43,
            }
        ],
        "enum_list": [
            "x",
            "y",
            "z",
        ],
        "some_tuple": [
            {
                "foo": "A-1",
                "bar": 101,
            },
            {
                "foo": "B-1",
                "bar": 201,
                "baz": ["a", "b", "c"],
                "qux": -100,
                "quux": 43,
            },
        ],
    }


@pytest.mark.unit
def test_instantiation_with_a_from_legacy_dict_method():
    """Can be instantiated from a legacy dictionary.

    Note: This pattern is helpful for cases where we're migrating from dictionary-based objects to typed objects.
    One especially thorny example is when the dictionary contains keys that are reserved words in python.

    For example, test cases use the reserved word: "in" as one of their required fields.
    """  # noqa: E501 # FIXME CoP

    import inspect
    import logging

    @dataclass
    class MyClassE(SerializableDictDot):
        foo: str
        bar: int
        input: int

        @classmethod
        def from_legacy_dict(cls, dict):
            """This method is an adapter to allow typing of legacy my_class_e dictionary objects, without needing to immediately clean up every object."""  # noqa: E501 # FIXME CoP
            temp_dict = {}
            for k, v in dict.items():
                # Ignore parameters that don't match the type definition
                if k in inspect.signature(cls).parameters:
                    temp_dict[k] = v
                else:  # noqa: PLR5501 # FIXME CoP
                    if k == "in":
                        temp_dict["input"] = v
                    else:
                        logging.warning(
                            f"WARNING: Got extra parameter: {k} while instantiating MyClassE."
                            "This parameter will be ignored."
                            "You probably need to clean up a library_metadata object."
                        )

            return cls(**temp_dict)

    my_E = MyClassE.from_legacy_dict(
        {
            "foo": "a string",
            "bar": 1,
            "in": 10,
        }
    )

    assert my_E["foo"] == "a string"
    assert my_E["bar"] == 1
    assert my_E["input"] == 10
    assert my_E.input == 10

    # Note that after instantiation, the class does NOT have an "in" property
    with raises(AttributeError):
        my_E["in"] == 10  # noqa: B015 # FIXME CoP

    # Because `in` is a reserved word, this will raise a SyntaxError:
    # my_F.in == 100

    # Because `in` is a reserved word, this will also raise a SyntaxError:
    # my_F.in = 100
